"""ANTs' utilities."""
import os
from warnings import warn
from ..base import traits, isdefined, TraitedSpec, File, Str, InputMultiObject
from ..mixins import CopyHeaderInterface
from .base import ANTSCommandInputSpec, ANTSCommand


class ImageMathInputSpec(ANTSCommandInputSpec):
    dimension = traits.Int(
        3, usedefault=True, position=1, argstr="%d", desc="dimension of output image"
    )
    output_image = File(
        position=2,
        argstr="%s",
        name_source=["op1"],
        name_template="%s_maths",
        desc="output image file",
        keep_extension=True,
    )
    operation = traits.Enum(
        "m",
        "vm",
        "+",
        "v+",
        "-",
        "v-",
        "/",
        "^",
        "max",
        "exp",
        "addtozero",
        "overadd",
        "abs",
        "total",
        "mean",
        "vtotal",
        "Decision",
        "Neg",
        "Project",
        "G",
        "MD",
        "ME",
        "MO",
        "MC",
        "GD",
        "GE",
        "GO",
        "GC",
        "TruncateImageIntensity",
        "Laplacian",
        "GetLargestComponent",
        "FillHoles",
        "PadImage",
        mandatory=True,
        position=3,
        argstr="%s",
        desc="mathematical operations",
    )
    op1 = File(
        exists=True, mandatory=True, position=-2, argstr="%s", desc="first operator"
    )
    op2 = traits.Either(
        File(exists=True), Str, position=-1, argstr="%s", desc="second operator"
    )
    copy_header = traits.Bool(
        True,
        usedefault=True,
        desc="copy headers of the original image into the output (corrected) file",
    )


class ImageMathOuputSpec(TraitedSpec):
    output_image = File(exists=True, desc="output image file")


class ImageMath(ANTSCommand, CopyHeaderInterface):
    """
    Operations over images.

    Examples
    --------
    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='+',
    ...     op2='2').cmdline
    'ImageMath 3 structural_maths.nii + structural.nii 2'

    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='Project',
    ...     op2='1 2').cmdline
    'ImageMath 3 structural_maths.nii Project structural.nii 1 2'

    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='G',
    ...     op2='4').cmdline
    'ImageMath 3 structural_maths.nii G structural.nii 4'

    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='TruncateImageIntensity',
    ...     op2='0.005 0.999 256').cmdline
    'ImageMath 3 structural_maths.nii TruncateImageIntensity structural.nii 0.005 0.999 256'

    By default, Nipype copies headers from the first input image (``op1``)
    to the output image.
    For the ``PadImage`` operation, the header cannot be copied from inputs to
    outputs, and so ``copy_header`` option is automatically set to ``False``.

    >>> pad = ImageMath(
    ...     op1='structural.nii',
    ...     operation='PadImage')
    >>> pad.inputs.copy_header
    False

    While the operation is set to ``PadImage``,
    setting ``copy_header = True`` will have no effect.

    >>> pad.inputs.copy_header = True
    >>> pad.inputs.copy_header
    False

    For any other operation, ``copy_header`` can be enabled/disabled normally:

    >>> pad.inputs.operation = "ME"
    >>> pad.inputs.copy_header = True
    >>> pad.inputs.copy_header
    True

    """

    _cmd = "ImageMath"
    input_spec = ImageMathInputSpec
    output_spec = ImageMathOuputSpec
    _copy_header_map = {"output_image": "op1"}

    def __init__(self, **inputs):
        super(ImageMath, self).__init__(**inputs)
        if self.inputs.operation in ("PadImage",):
            self.inputs.copy_header = False

        self.inputs.on_trait_change(self._operation_update, "operation")
        self.inputs.on_trait_change(self._copyheader_update, "copy_header")

    def _operation_update(self):
        if self.inputs.operation in ("PadImage",):
            self.inputs.copy_header = False

    def _copyheader_update(self):
        if self.inputs.copy_header and self.inputs.operation in ("PadImage",):
            warn("copy_header cannot be updated to True with PadImage as operation.")
            self.inputs.copy_header = False


class ResampleImageBySpacingInputSpec(ANTSCommandInputSpec):
    dimension = traits.Int(
        3, usedefault=True, position=1, argstr="%d", desc="dimension of output image"
    )
    input_image = File(
        exists=True, mandatory=True, position=2, argstr="%s", desc="input image file"
    )
    output_image = File(
        position=3,
        argstr="%s",
        name_source=["input_image"],
        name_template="%s_resampled",
        desc="output image file",
        keep_extension=True,
    )
    out_spacing = traits.Either(
        traits.List(traits.Float, minlen=2, maxlen=3),
        traits.Tuple(traits.Float, traits.Float, traits.Float),
        traits.Tuple(traits.Float, traits.Float),
        position=4,
        argstr="%s",
        mandatory=True,
        desc="output spacing",
    )
    apply_smoothing = traits.Bool(
        False, argstr="%d", position=5, desc="smooth before resampling"
    )
    addvox = traits.Int(
        argstr="%d",
        position=6,
        requires=["apply_smoothing"],
        desc="addvox pads each dimension by addvox",
    )
    nn_interp = traits.Bool(
        argstr="%d", desc="nn interpolation", position=-1, requires=["addvox"]
    )


class ResampleImageBySpacingOutputSpec(TraitedSpec):
    output_image = File(exists=True, desc="resampled file")


class ResampleImageBySpacing(ANTSCommand):
    """
    Resample an image with a given spacing.

    Examples
    --------
    >>> res = ResampleImageBySpacing(dimension=3)
    >>> res.inputs.input_image = 'structural.nii'
    >>> res.inputs.output_image = 'output.nii.gz'
    >>> res.inputs.out_spacing = (4, 4, 4)
    >>> res.cmdline  #doctest: +ELLIPSIS
    'ResampleImageBySpacing 3 structural.nii output.nii.gz 4 4 4'

    >>> res = ResampleImageBySpacing(dimension=3)
    >>> res.inputs.input_image = 'structural.nii'
    >>> res.inputs.output_image = 'output.nii.gz'
    >>> res.inputs.out_spacing = (4, 4, 4)
    >>> res.inputs.apply_smoothing = True
    >>> res.cmdline  #doctest: +ELLIPSIS
    'ResampleImageBySpacing 3 structural.nii output.nii.gz 4 4 4 1'

    >>> res = ResampleImageBySpacing(dimension=3)
    >>> res.inputs.input_image = 'structural.nii'
    >>> res.inputs.output_image = 'output.nii.gz'
    >>> res.inputs.out_spacing = (0.4, 0.4, 0.4)
    >>> res.inputs.apply_smoothing = True
    >>> res.inputs.addvox = 2
    >>> res.inputs.nn_interp = False
    >>> res.cmdline  #doctest: +ELLIPSIS
    'ResampleImageBySpacing 3 structural.nii output.nii.gz 0.4 0.4 0.4 1 2 0'

    """

    _cmd = "ResampleImageBySpacing"
    input_spec = ResampleImageBySpacingInputSpec
    output_spec = ResampleImageBySpacingOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if name == "out_spacing":
            if len(value) != self.inputs.dimension:
                raise ValueError("out_spacing dimensions should match dimension")

            value = " ".join(["%g" % d for d in value])

        return super(ResampleImageBySpacing, self)._format_arg(name, trait_spec, value)


class ThresholdImageInputSpec(ANTSCommandInputSpec):
    dimension = traits.Int(
        3, usedefault=True, position=1, argstr="%d", desc="dimension of output image"
    )
    input_image = File(
        exists=True, mandatory=True, position=2, argstr="%s", desc="input image file"
    )
    output_image = File(
        position=3,
        argstr="%s",
        name_source=["input_image"],
        name_template="%s_resampled",
        desc="output image file",
        keep_extension=True,
    )

    mode = traits.Enum(
        "Otsu",
        "Kmeans",
        argstr="%s",
        position=4,
        requires=["num_thresholds"],
        xor=["th_low", "th_high"],
        desc="whether to run Otsu / Kmeans thresholding",
    )
    num_thresholds = traits.Int(position=5, argstr="%d", desc="number of thresholds")
    input_mask = File(
        exists=True,
        requires=["num_thresholds"],
        argstr="%s",
        desc="input mask for Otsu, Kmeans",
    )

    th_low = traits.Float(position=4, argstr="%f", xor=["mode"], desc="lower threshold")
    th_high = traits.Float(
        position=5, argstr="%f", xor=["mode"], desc="upper threshold"
    )
    inside_value = traits.Float(
        1, position=6, argstr="%f", requires=["th_low"], desc="inside value"
    )
    outside_value = traits.Float(
        0, position=7, argstr="%f", requires=["th_low"], desc="outside value"
    )
    copy_header = traits.Bool(
        True,
        mandatory=True,
        usedefault=True,
        desc="copy headers of the original image into the output (corrected) file",
    )


class ThresholdImageOutputSpec(TraitedSpec):
    output_image = File(exists=True, desc="resampled file")


class ThresholdImage(ANTSCommand, CopyHeaderInterface):
    """
    Apply thresholds on images.

    Examples
    --------
    >>> thres = ThresholdImage(dimension=3)
    >>> thres.inputs.input_image = 'structural.nii'
    >>> thres.inputs.output_image = 'output.nii.gz'
    >>> thres.inputs.th_low = 0.5
    >>> thres.inputs.th_high = 1.0
    >>> thres.inputs.inside_value = 1.0
    >>> thres.inputs.outside_value = 0.0
    >>> thres.cmdline  #doctest: +ELLIPSIS
    'ThresholdImage 3 structural.nii output.nii.gz 0.500000 1.000000 1.000000 0.000000'

    >>> thres = ThresholdImage(dimension=3)
    >>> thres.inputs.input_image = 'structural.nii'
    >>> thres.inputs.output_image = 'output.nii.gz'
    >>> thres.inputs.mode = 'Kmeans'
    >>> thres.inputs.num_thresholds = 4
    >>> thres.cmdline  #doctest: +ELLIPSIS
    'ThresholdImage 3 structural.nii output.nii.gz Kmeans 4'

    """

    _cmd = "ThresholdImage"
    input_spec = ThresholdImageInputSpec
    output_spec = ThresholdImageOutputSpec
    _copy_header_map = {"output_image": "input_image"}


class AIInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, usedefault=True, argstr="-d %d", desc="dimension of output image"
    )
    verbose = traits.Bool(
        False, usedefault=True, argstr="-v %d", desc="enable verbosity"
    )

    fixed_image = File(
        exists=True,
        mandatory=True,
        desc="Image to which the moving_image should be transformed",
    )
    moving_image = File(
        exists=True,
        mandatory=True,
        desc="Image that will be transformed to fixed_image",
    )

    fixed_image_mask = File(exists=True, argstr="-x %s", desc="fixed mage mask")
    moving_image_mask = File(
        exists=True, requires=["fixed_image_mask"], desc="moving mage mask"
    )

    metric_trait = (
        traits.Enum("Mattes", "GC", "MI"),
        traits.Int(32),
        traits.Enum("Regular", "Random", "None"),
        traits.Range(value=0.2, low=0.0, high=1.0),
    )
    metric = traits.Tuple(
        *metric_trait, argstr="-m %s", mandatory=True, desc="the metric(s) to use."
    )

    transform = traits.Tuple(
        traits.Enum("Affine", "Rigid", "Similarity"),
        traits.Range(value=0.1, low=0.0, exclude_low=True),
        argstr="-t %s[%g]",
        usedefault=True,
        desc="Several transform options are available",
    )

    principal_axes = traits.Bool(
        False,
        usedefault=True,
        argstr="-p %d",
        xor=["blobs"],
        desc="align using principal axes",
    )
    search_factor = traits.Tuple(
        traits.Float(20),
        traits.Range(value=0.12, low=0.0, high=1.0),
        usedefault=True,
        argstr="-s [%g,%g]",
        desc="search factor",
    )

    search_grid = traits.Either(
        traits.Tuple(
            traits.Float, traits.Tuple(traits.Float, traits.Float, traits.Float)
        ),
        traits.Tuple(traits.Float, traits.Tuple(traits.Float, traits.Float)),
        argstr="-g %s",
        desc="Translation search grid in mm",
        min_ver="2.3.0",
    )

    convergence = traits.Tuple(
        traits.Range(low=1, high=10000, value=10),
        traits.Float(1e-6),
        traits.Range(low=1, high=100, value=10),
        usedefault=True,
        argstr="-c [%d,%g,%d]",
        desc="convergence",
    )

    output_transform = File(
        "initialization.mat", usedefault=True, argstr="-o %s", desc="output file name"
    )


class AIOuputSpec(TraitedSpec):
    output_transform = File(exists=True, desc="output file name")


class AI(ANTSCommand):
    """
    Calculate the optimal linear transform parameters for aligning two images.

    Examples
    --------
    >>> AI(
    ...     fixed_image='structural.nii',
    ...     moving_image='epi.nii',
    ...     metric=('Mattes', 32, 'Regular', 1),
    ... ).cmdline
    'antsAI -c [10,1e-06,10] -d 3 -m Mattes[structural.nii,epi.nii,32,Regular,1]
    -o initialization.mat -p 0 -s [20,0.12] -t Affine[0.1] -v 0'

    >>> AI(fixed_image='structural.nii',
    ...    moving_image='epi.nii',
    ...    metric=('Mattes', 32, 'Regular', 1),
    ...    search_grid=(12, (1, 1, 1)),
    ... ).cmdline
    'antsAI -c [10,1e-06,10] -d 3 -m Mattes[structural.nii,epi.nii,32,Regular,1]
    -o initialization.mat -p 0 -s [20,0.12] -g [12.0,1x1x1] -t Affine[0.1] -v 0'

    """

    _cmd = "antsAI"
    input_spec = AIInputSpec
    output_spec = AIOuputSpec

    def _run_interface(self, runtime, correct_return_codes=(0,)):
        runtime = super(AI, self)._run_interface(runtime, correct_return_codes)

        self._output = {
            "output_transform": os.path.join(
                runtime.cwd, os.path.basename(self.inputs.output_transform)
            )
        }
        return runtime

    def _format_arg(self, opt, spec, val):
        if opt == "metric":
            val = "%s[{fixed_image},{moving_image},%d,%s,%g]" % val
            val = val.format(
                fixed_image=self.inputs.fixed_image,
                moving_image=self.inputs.moving_image,
            )
            return spec.argstr % val

        if opt == "search_grid":
            fmtval = "[%s,%s]" % (val[0], "x".join("%g" % v for v in val[1]))
            return spec.argstr % fmtval

        if opt == "fixed_image_mask":
            if isdefined(self.inputs.moving_image_mask):
                return spec.argstr % ("[%s,%s]" % (val, self.inputs.moving_image_mask))

        return super(AI, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        return getattr(self, "_output")


class AverageAffineTransformInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    output_affine_transform = File(
        argstr="%s",
        mandatory=True,
        position=1,
        desc="Outputfname.txt: the name of the resulting transform.",
    )
    transforms = InputMultiObject(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=3,
        desc="transforms to average",
    )


class AverageAffineTransformOutputSpec(TraitedSpec):
    affine_transform = File(exists=True, desc="average transform file")


class AverageAffineTransform(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import AverageAffineTransform
    >>> avg = AverageAffineTransform()
    >>> avg.inputs.dimension = 3
    >>> avg.inputs.transforms = ['trans.mat', 'func_to_struct.mat']
    >>> avg.inputs.output_affine_transform = 'MYtemplatewarp.mat'
    >>> avg.cmdline
    'AverageAffineTransform 3 MYtemplatewarp.mat trans.mat func_to_struct.mat'

    """

    _cmd = "AverageAffineTransform"
    input_spec = AverageAffineTransformInputSpec
    output_spec = AverageAffineTransformOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(AverageAffineTransform, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["affine_transform"] = os.path.abspath(
            self.inputs.output_affine_transform
        )
        return outputs


class AverageImagesInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    output_average_image = File(
        "average.nii",
        argstr="%s",
        position=1,
        usedefault=True,
        hash_files=False,
        desc="the name of the resulting image.",
    )
    normalize = traits.Bool(
        argstr="%d",
        mandatory=True,
        position=2,
        desc="Normalize: if true, the 2nd image is divided by its mean. "
        "This will select the largest image to average into.",
    )
    images = InputMultiObject(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=3,
        desc="image to apply transformation to (generally a coregistered functional)",
    )


class AverageImagesOutputSpec(TraitedSpec):
    output_average_image = File(exists=True, desc="average image file")


class AverageImages(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import AverageImages
    >>> avg = AverageImages()
    >>> avg.inputs.dimension = 3
    >>> avg.inputs.output_average_image = "average.nii.gz"
    >>> avg.inputs.normalize = True
    >>> avg.inputs.images = ['rc1s1.nii', 'rc1s1.nii']
    >>> avg.cmdline
    'AverageImages 3 average.nii.gz 1 rc1s1.nii rc1s1.nii'
    """

    _cmd = "AverageImages"
    input_spec = AverageImagesInputSpec
    output_spec = AverageImagesOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(AverageImages, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["output_average_image"] = os.path.realpath(
            self.inputs.output_average_image
        )
        return outputs


class MultiplyImagesInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    first_input = File(
        argstr="%s", exists=True, mandatory=True, position=1, desc="image 1"
    )
    second_input = traits.Either(
        File(exists=True),
        traits.Float,
        argstr="%s",
        mandatory=True,
        position=2,
        desc="image 2 or multiplication weight",
    )
    output_product_image = File(
        argstr="%s",
        mandatory=True,
        position=3,
        desc="Outputfname.nii.gz: the name of the resulting image.",
    )


class MultiplyImagesOutputSpec(TraitedSpec):
    output_product_image = File(exists=True, desc="average image file")


class MultiplyImages(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import MultiplyImages
    >>> test = MultiplyImages()
    >>> test.inputs.dimension = 3
    >>> test.inputs.first_input = 'moving2.nii'
    >>> test.inputs.second_input = 0.25
    >>> test.inputs.output_product_image = "out.nii"
    >>> test.cmdline
    'MultiplyImages 3 moving2.nii 0.25 out.nii'
    """

    _cmd = "MultiplyImages"
    input_spec = MultiplyImagesInputSpec
    output_spec = MultiplyImagesOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(MultiplyImages, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["output_product_image"] = os.path.abspath(
            self.inputs.output_product_image
        )
        return outputs


class CreateJacobianDeterminantImageInputSpec(ANTSCommandInputSpec):
    imageDimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    deformationField = File(
        argstr="%s",
        exists=True,
        mandatory=True,
        position=1,
        desc="deformation transformation file",
    )
    outputImage = File(argstr="%s", mandatory=True, position=2, desc="output filename")
    doLogJacobian = traits.Enum(
        0, 1, argstr="%d", position=3, desc="return the log jacobian"
    )
    useGeometric = traits.Enum(
        0, 1, argstr="%d", position=4, desc="return the geometric jacobian"
    )


class CreateJacobianDeterminantImageOutputSpec(TraitedSpec):
    jacobian_image = File(exists=True, desc="jacobian image")


class CreateJacobianDeterminantImage(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import CreateJacobianDeterminantImage
    >>> jacobian = CreateJacobianDeterminantImage()
    >>> jacobian.inputs.imageDimension = 3
    >>> jacobian.inputs.deformationField = 'ants_Warp.nii.gz'
    >>> jacobian.inputs.outputImage = 'out_name.nii.gz'
    >>> jacobian.cmdline
    'CreateJacobianDeterminantImage 3 ants_Warp.nii.gz out_name.nii.gz'
    """

    _cmd = "CreateJacobianDeterminantImage"
    input_spec = CreateJacobianDeterminantImageInputSpec
    output_spec = CreateJacobianDeterminantImageOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(CreateJacobianDeterminantImage, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["jacobian_image"] = os.path.abspath(self.inputs.outputImage)
        return outputs


class AffineInitializerInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, usedefault=True, position=0, argstr="%s", desc="dimension"
    )
    fixed_image = File(
        exists=True, mandatory=True, position=1, argstr="%s", desc="reference image"
    )
    moving_image = File(
        exists=True, mandatory=True, position=2, argstr="%s", desc="moving image"
    )
    out_file = File(
        "transform.mat",
        usedefault=True,
        position=3,
        argstr="%s",
        desc="output transform file",
    )
    # Defaults in antsBrainExtraction.sh -> 15 0.1 0 10
    search_factor = traits.Float(
        15.0,
        usedefault=True,
        position=4,
        argstr="%f",
        desc="increments (degrees) for affine search",
    )
    radian_fraction = traits.Range(
        0.0,
        1.0,
        value=0.1,
        usedefault=True,
        position=5,
        argstr="%f",
        desc="search this arc +/- principal axes",
    )
    principal_axes = traits.Bool(
        False,
        usedefault=True,
        position=6,
        argstr="%d",
        desc="whether the rotation is searched around an initial principal axis alignment.",
    )
    local_search = traits.Int(
        10,
        usedefault=True,
        position=7,
        argstr="%d",
        desc=" determines if a local optimization is run at each search point for the set "
        "number of iterations",
    )


class AffineInitializerOutputSpec(TraitedSpec):
    out_file = File(desc="output transform file")


class AffineInitializer(ANTSCommand):
    """
    Initialize an affine transform (as in antsBrainExtraction.sh)

    >>> from nipype.interfaces.ants import AffineInitializer
    >>> init = AffineInitializer()
    >>> init.inputs.fixed_image = 'fixed1.nii'
    >>> init.inputs.moving_image = 'moving1.nii'
    >>> init.cmdline
    'antsAffineInitializer 3 fixed1.nii moving1.nii transform.mat 15.000000 0.100000 0 10'

    """

    _cmd = "antsAffineInitializer"
    input_spec = AffineInitializerInputSpec
    output_spec = AffineInitializerOutputSpec

    def _list_outputs(self):
        return {"out_file": os.path.abspath(self.inputs.out_file)}


class ComposeMultiTransformInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", usedefault=True, position=0, desc="image dimension (2 or 3)"
    )
    output_transform = File(
        argstr="%s",
        position=1,
        name_source=["transforms"],
        name_template="%s_composed",
        keep_extension=True,
        desc="the name of the resulting transform.",
    )
    reference_image = File(
        argstr="%s",
        position=2,
        desc="Reference image (only necessary when output is warpfield)",
    )
    transforms = InputMultiObject(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=3,
        desc="transforms to average",
    )


class ComposeMultiTransformOutputSpec(TraitedSpec):
    output_transform = File(exists=True, desc="Composed transform file")


class ComposeMultiTransform(ANTSCommand):
    """
    Take a set of transformations and convert them to a single transformation matrix/warpfield.

    Examples
    --------
    >>> from nipype.interfaces.ants import ComposeMultiTransform
    >>> compose_transform = ComposeMultiTransform()
    >>> compose_transform.inputs.dimension = 3
    >>> compose_transform.inputs.transforms = ['struct_to_template.mat', 'func_to_struct.mat']
    >>> compose_transform.cmdline
    'ComposeMultiTransform 3 struct_to_template_composed.mat
    struct_to_template.mat func_to_struct.mat'

    """

    _cmd = "ComposeMultiTransform"
    input_spec = ComposeMultiTransformInputSpec
    output_spec = ComposeMultiTransformOutputSpec


class LabelGeometryInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", usedefault=True, position=0, desc="image dimension (2 or 3)"
    )
    label_image = File(
        argstr="%s",
        position=1,
        mandatory=True,
        desc="label image to use for extracting geometry measures",
    )
    intensity_image = File(
        value="[]",
        exists=True,
        argstr="%s",
        mandatory=True,
        usedefault=True,
        position=2,
        desc="Intensity image to extract values from. " "This is an optional input",
    )
    output_file = traits.Str(
        name_source=["label_image"],
        name_template="%s.csv",
        argstr="%s",
        position=3,
        desc="name of output file",
    )


class LabelGeometryOutputSpec(TraitedSpec):
    output_file = File(exists=True, desc="CSV file of geometry measures")


class LabelGeometry(ANTSCommand):
    """
    Extracts geometry measures using a label file and an optional image file

    Examples
    --------
    >>> from nipype.interfaces.ants import LabelGeometry
    >>> label_extract = LabelGeometry()
    >>> label_extract.inputs.dimension = 3
    >>> label_extract.inputs.label_image = 'atlas.nii.gz'
    >>> label_extract.cmdline
    'LabelGeometryMeasures 3 atlas.nii.gz [] atlas.csv'

    >>> label_extract.inputs.intensity_image = 'ants_Warp.nii.gz'
    >>> label_extract.cmdline
    'LabelGeometryMeasures 3 atlas.nii.gz ants_Warp.nii.gz atlas.csv'

    """

    _cmd = "LabelGeometryMeasures"
    input_spec = LabelGeometryInputSpec
    output_spec = LabelGeometryOutputSpec



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os.path as op
import nibabel as nb
import nibabel.trackvis as trk
import numpy as np
from nibabel.trackvis import HeaderError
from nibabel.volumeutils import native_code
from nibabel.orientations import aff2axcodes

from ... import logging
from ...utils.filemanip import split_filename
from ..base import TraitedSpec, File, isdefined
from ..dipy.base import DipyBaseInterface, HAVE_DIPY as have_dipy

iflogger = logging.getLogger("nipype.interface")


def get_vox_dims(volume):
    import nibabel as nb

    if isinstance(volume, list):
        volume = volume[0]
    nii = nb.load(volume)
    hdr = nii.header
    voxdims = hdr.get_zooms()
    return [float(voxdims[0]), float(voxdims[1]), float(voxdims[2])]


def get_data_dims(volume):
    import nibabel as nb

    if isinstance(volume, list):
        volume = volume[0]
    nii = nb.load(volume)
    hdr = nii.header
    datadims = hdr.get_data_shape()
    return [int(datadims[0]), int(datadims[1]), int(datadims[2])]


def transform_to_affine(streams, header, affine):
    from dipy.tracking.utils import move_streamlines

    rotation, scale = np.linalg.qr(affine)
    streams = move_streamlines(streams, rotation)
    scale[0:3, 0:3] = np.dot(scale[0:3, 0:3], np.diag(1.0 / header["voxel_size"]))
    scale[0:3, 3] = abs(scale[0:3, 3])
    streams = move_streamlines(streams, scale)
    return streams


def read_mrtrix_tracks(in_file, as_generator=True):
    header = read_mrtrix_header(in_file)
    streamlines = read_mrtrix_streamlines(in_file, header, as_generator)
    return header, streamlines


def read_mrtrix_header(in_file):
    fileobj = open(in_file, "rb")
    header = {}
    iflogger.info("Reading header data...")
    for line in fileobj:
        line = line.decode()
        if line == "END\n":
            iflogger.info("Reached the end of the header!")
            break
        elif ": " in line:
            line = line.replace("\n", "")
            line = line.replace("'", "")
            key = line.split(": ")[0]
            value = line.split(": ")[1]
            header[key] = value
            iflogger.info('...adding "%s" to header for key "%s"', value, key)
    fileobj.close()
    header["count"] = int(header["count"].replace("\n", ""))
    header["offset"] = int(header["file"].replace(".", ""))
    return header


def read_mrtrix_streamlines(in_file, header, as_generator=True):
    offset = header["offset"]
    stream_count = header["count"]
    fileobj = open(in_file, "rb")
    fileobj.seek(offset)
    endianness = native_code
    f4dt = np.dtype(endianness + "f4")
    pt_cols = 3
    bytesize = pt_cols * 4

    def points_per_track(offset):
        track_points = []
        iflogger.info("Identifying the number of points per tract...")
        all_str = fileobj.read()
        num_triplets = int(len(all_str) / bytesize)
        pts = np.ndarray(shape=(num_triplets, pt_cols), dtype="f4", buffer=all_str)
        nonfinite_list = np.where(np.invert(np.isfinite(pts[:, 2])))
        nonfinite_list = list(nonfinite_list[0])[
            0:-1
        ]  # Converts numpy array to list, removes the last value
        for idx, value in enumerate(nonfinite_list):
            if idx == 0:
                track_points.append(nonfinite_list[idx])
            else:
                track_points.append(nonfinite_list[idx] - nonfinite_list[idx - 1] - 1)
        return track_points, nonfinite_list

    def track_gen(track_points):
        n_streams = 0
        iflogger.info("Reading tracks...")
        while True:
            try:
                n_pts = track_points[n_streams]
            except IndexError:
                break
            pts_str = fileobj.read(n_pts * bytesize)
            nan_str = fileobj.read(bytesize)
            if len(pts_str) < (n_pts * bytesize):
                if not n_streams == stream_count:
                    raise HeaderError(
                        "Expecting %s points, found only %s" % (stream_count, n_streams)
                    )
                    iflogger.error(
                        "Expecting %s points, found only %s", stream_count, n_streams
                    )
                break
            pts = np.ndarray(shape=(n_pts, pt_cols), dtype=f4dt, buffer=pts_str)
            nan_pt = np.ndarray(shape=(1, pt_cols), dtype=f4dt, buffer=nan_str)
            if np.isfinite(nan_pt[0][0]):
                raise ValueError
                break
            xyz = pts[:, :3]
            yield xyz
            n_streams += 1
            if n_streams == stream_count:
                iflogger.info("100%% : %i tracks read", n_streams)
                raise StopIteration
            try:
                if n_streams % int(stream_count / 100) == 0:
                    percent = int(float(n_streams) / float(stream_count) * 100)
                    iflogger.info("%i%% : %i tracks read", percent, n_streams)
            except ZeroDivisionError:
                iflogger.info("%i stream read out of %i", n_streams, stream_count)

    track_points, nonfinite_list = points_per_track(offset)
    fileobj.seek(offset)
    streamlines = track_gen(track_points)
    if not as_generator:
        streamlines = list(streamlines)
    return streamlines


class MRTrix2TrackVisInputSpec(TraitedSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        desc="The input file for the tracks in MRTrix (.tck) format",
    )
    image_file = File(exists=True, desc="The image the tracks were generated from")
    matrix_file = File(
        exists=True,
        desc="A transformation matrix to apply to the tracts after they have been generated (from FLIRT - affine transformation from image_file to registration_image_file)",
    )
    registration_image_file = File(
        exists=True, desc="The final image the tracks should be registered to."
    )
    out_filename = File(
        "converted.trk",
        genfile=True,
        usedefault=True,
        desc="The output filename for the tracks in TrackVis (.trk) format",
    )


class MRTrix2TrackVisOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class MRTrix2TrackVis(DipyBaseInterface):
    """
    Converts MRtrix (.tck) tract files into TrackVis (.trk) format
    using functions from dipy
    Example
    -------
    >>> import nipype.interfaces.mrtrix as mrt
    >>> tck2trk = mrt.MRTrix2TrackVis()
    >>> tck2trk.inputs.in_file = 'dwi_CSD_tracked.tck'
    >>> tck2trk.inputs.image_file = 'diffusion.nii'
    >>> tck2trk.run()                                   # doctest: +SKIP
    """

    input_spec = MRTrix2TrackVisInputSpec
    output_spec = MRTrix2TrackVisOutputSpec

    def _run_interface(self, runtime):
        from dipy.tracking.utils import move_streamlines, affine_from_fsl_mat_file

        dx, dy, dz = get_data_dims(self.inputs.image_file)
        vx, vy, vz = get_vox_dims(self.inputs.image_file)
        image_file = nb.load(self.inputs.image_file)
        affine = image_file.affine
        out_filename = op.abspath(self.inputs.out_filename)

        # Reads MRTrix tracks
        header, streamlines = read_mrtrix_tracks(self.inputs.in_file, as_generator=True)
        iflogger.info("MRTrix Header:")
        iflogger.info(header)
        # Writes to Trackvis
        trk_header = nb.trackvis.empty_header()
        trk_header["dim"] = [dx, dy, dz]
        trk_header["voxel_size"] = [vx, vy, vz]
        trk_header["n_count"] = header["count"]

        if isdefined(self.inputs.matrix_file) and isdefined(
            self.inputs.registration_image_file
        ):
            iflogger.info(
                "Applying transformation from matrix file %s", self.inputs.matrix_file
            )
            xfm = np.genfromtxt(self.inputs.matrix_file)
            iflogger.info(xfm)
            registration_image_file = nb.load(self.inputs.registration_image_file)
            reg_affine = registration_image_file.affine
            r_dx, r_dy, r_dz = get_data_dims(self.inputs.registration_image_file)
            r_vx, r_vy, r_vz = get_vox_dims(self.inputs.registration_image_file)
            iflogger.info(
                "Using affine from registration image file %s",
                self.inputs.registration_image_file,
            )
            iflogger.info(reg_affine)
            trk_header["vox_to_ras"] = reg_affine
            trk_header["dim"] = [r_dx, r_dy, r_dz]
            trk_header["voxel_size"] = [r_vx, r_vy, r_vz]

            affine = np.dot(affine, np.diag(1.0 / np.array([vx, vy, vz, 1])))
            transformed_streamlines = transform_to_affine(
                streamlines, trk_header, affine
            )

            aff = affine_from_fsl_mat_file(xfm, [vx, vy, vz], [r_vx, r_vy, r_vz])
            iflogger.info(aff)

            axcode = aff2axcodes(reg_affine)
            trk_header["voxel_order"] = axcode[0] + axcode[1] + axcode[2]

            final_streamlines = move_streamlines(transformed_streamlines, aff)
            trk_tracks = ((ii, None, None) for ii in final_streamlines)
            trk.write(out_filename, trk_tracks, trk_header)
            iflogger.info("Saving transformed Trackvis file as %s", out_filename)
            iflogger.info("New TrackVis Header:")
            iflogger.info(trk_header)
        else:
            iflogger.info(
                "Applying transformation from scanner coordinates to %s",
                self.inputs.image_file,
            )
            axcode = aff2axcodes(affine)
            trk_header["voxel_order"] = axcode[0] + axcode[1] + axcode[2]
            trk_header["vox_to_ras"] = affine
            transformed_streamlines = transform_to_affine(
                streamlines, trk_header, affine
            )
            trk_tracks = ((ii, None, None) for ii in transformed_streamlines)
            trk.write(out_filename, trk_tracks, trk_header)
            iflogger.info("Saving Trackvis file as %s", out_filename)
            iflogger.info("TrackVis Header:")
            iflogger.info(trk_header)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = op.abspath(self.inputs.out_filename)
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + ".trk"



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import Concatenate


def test_Concatenate_inputs():
    input_map = dict(
        add_val=dict(
            argstr="--add %f",
        ),
        args=dict(
            argstr="%s",
        ),
        combine=dict(
            argstr="--combine",
        ),
        concatenated_file=dict(
            argstr="--o %s",
            extensions=None,
            genfile=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        gmean=dict(
            argstr="--gmean %d",
        ),
        in_files=dict(
            argstr="--i %s...",
            mandatory=True,
        ),
        keep_dtype=dict(
            argstr="--keep-datatype",
        ),
        mask_file=dict(
            argstr="--mask %s",
            extensions=None,
        ),
        max_bonfcor=dict(
            argstr="--max-bonfcor",
        ),
        max_index=dict(
            argstr="--max-index",
        ),
        mean_div_n=dict(
            argstr="--mean-div-n",
        ),
        multiply_by=dict(
            argstr="--mul %f",
        ),
        multiply_matrix_file=dict(
            argstr="--mtx %s",
            extensions=None,
        ),
        paired_stats=dict(
            argstr="--paired-%s",
        ),
        sign=dict(
            argstr="--%s",
        ),
        sort=dict(
            argstr="--sort",
        ),
        stats=dict(
            argstr="--%s",
        ),
        subjects_dir=dict(),
        vote=dict(
            argstr="--vote",
        ),
    )
    inputs = Concatenate.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Concatenate_outputs():
    output_map = dict(
        concatenated_file=dict(
            extensions=None,
        ),
    )
    outputs = Concatenate.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brainsuite import Pialmesh


def test_Pialmesh_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        exportPrefix=dict(
            argstr="--prefix %s",
        ),
        inputMaskFile=dict(
            argstr="-m %s",
            extensions=None,
            mandatory=True,
        ),
        inputSurfaceFile=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
        ),
        inputTissueFractionFile=dict(
            argstr="-f %s",
            extensions=None,
            mandatory=True,
        ),
        laplacianSmoothing=dict(
            argstr="--smooth %f",
            usedefault=True,
        ),
        maxThickness=dict(
            argstr="--max %f",
            usedefault=True,
        ),
        normalSmoother=dict(
            argstr="--nc %f",
            usedefault=True,
        ),
        numIterations=dict(
            argstr="-n %d",
            usedefault=True,
        ),
        outputInterval=dict(
            argstr="--interval %d",
            usedefault=True,
        ),
        outputSurfaceFile=dict(
            argstr="-o %s",
            extensions=None,
            genfile=True,
        ),
        recomputeNormals=dict(
            argstr="--norm",
        ),
        searchRadius=dict(
            argstr="-r %f",
            usedefault=True,
        ),
        stepSize=dict(
            argstr="-s %f",
            usedefault=True,
        ),
        tangentSmoother=dict(
            argstr="--tc %f",
        ),
        timer=dict(
            argstr="--timer",
        ),
        tissueThreshold=dict(
            argstr="-t %f",
            usedefault=True,
        ),
        verbosity=dict(
            argstr="-v %d",
        ),
    )
    inputs = Pialmesh.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Pialmesh_outputs():
    output_map = dict(
        outputSurfaceFile=dict(
            extensions=None,
        ),
    )
    outputs = Pialmesh.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The ASL module of niftyfit, which wraps the fitting methods in NiftyFit.
"""

from ..base import File, TraitedSpec, traits, CommandLineInputSpec
from .base import NiftyFitCommand
from ..niftyreg.base import get_custom_path


class FitAslInputSpec(CommandLineInputSpec):
    """Input Spec for FitAsl."""

    desc = "Filename of the 4D ASL (control/label) source image (mandatory)."
    source_file = File(
        position=1, exists=True, argstr="-source %s", mandatory=True, desc=desc
    )
    pasl = traits.Bool(desc="Fit PASL ASL data [default]", argstr="-pasl")
    pcasl = traits.Bool(desc="Fit PCASL ASL data", argstr="-pcasl")

    # *** Output options:
    desc = "Filename of the Cerebral Blood Flow map (in ml/100g/min)."
    cbf_file = File(
        name_source=["source_file"],
        name_template="%s_cbf.nii.gz",
        argstr="-cbf %s",
        desc=desc,
    )
    error_file = File(
        name_source=["source_file"],
        name_template="%s_error.nii.gz",
        argstr="-error %s",
        desc="Filename of the CBF error map.",
    )
    syn_file = File(
        name_source=["source_file"],
        name_template="%s_syn.nii.gz",
        argstr="-syn %s",
        desc="Filename of the synthetic ASL data.",
    )

    # *** Input options (see also fit_qt1 for generic T1 fitting):
    desc = "Filename of the estimated input T1 map (in ms)."
    t1map = File(exists=True, argstr="-t1map %s", desc=desc)
    desc = "Filename of the estimated input M0 map."
    m0map = File(exists=True, argstr="-m0map %s", desc=desc)
    desc = "Filename of the estimated input M0 map error."
    m0mape = File(exists=True, argstr="-m0mape %s", desc=desc)
    desc = "Filename of a [1,2,5]s Inversion Recovery volume (T1/M0 fitting \
carried out internally)."

    ir_volume = File(exists=True, argstr="-IRvolume %s", desc=desc)
    desc = "Output of [1,2,5]s Inversion Recovery fitting."
    ir_output = File(exists=True, argstr="-IRoutput %s", desc=desc)

    # *** Experimental options (Choose those suitable for the model!):
    mask = File(
        position=2, exists=True, desc="Filename of image mask.", argstr="-mask %s"
    )
    t1_art_cmp = traits.Float(
        desc="T1 of arterial component [1650ms].", argstr="-T1a %f"
    )
    desc = "Single plasma/tissue partition coefficient [0.9ml/g]."
    plasma_coeff = traits.Float(desc=desc, argstr="-L %f")
    desc = "Labelling efficiency [0.99 (pasl), 0.85 (pcasl)], ensure any \
background suppression pulses are included in -eff"

    eff = traits.Float(desc=desc, argstr="-eff %f")
    desc = "Outlier rejection for multi CL volumes (enter z-score threshold \
(e.g. 2.5)) [off]."

    out = traits.Float(desc=desc, argstr="-out %f")

    # *** PCASL options (Choose those suitable for the model!):
    pld = traits.Float(desc="Post Labelling Delay [2000ms].", argstr="-PLD %f")
    ldd = traits.Float(desc="Labelling Duration [1800ms].", argstr="-LDD %f")
    desc = "Difference in labelling delay per slice [0.0 ms/slice."
    dpld = traits.Float(desc=desc, argstr="-dPLD %f")

    # *** PASL options (Choose those suitable for the model!):
    t_inv1 = traits.Float(desc="Saturation pulse time [800ms].", argstr="-Tinv1 %f")
    t_inv2 = traits.Float(desc="Inversion time [2000ms].", argstr="-Tinv2 %f")
    desc = "Difference in inversion time per slice [0ms/slice]."
    dt_inv2 = traits.Float(desc=desc, argstr="-dTinv2 %f")

    # *** Other experimental assumptions:

    # Not programmed yet
    # desc = 'Slope and intercept for Arterial Transit Time.'
    # ATT = traits.Float(desc=desc, argstr='-ATT %f')

    gm_t1 = traits.Float(desc="T1 of GM [1150ms].", argstr="-gmT1 %f")
    gm_plasma = traits.Float(
        desc="Plasma/GM water partition [0.95ml/g].", argstr="-gmL %f"
    )
    gm_ttt = traits.Float(desc="Time to GM [ATT+0ms].", argstr="-gmTTT %f")
    wm_t1 = traits.Float(desc="T1 of WM [800ms].", argstr="-wmT1 %f")
    wm_plasma = traits.Float(
        desc="Plasma/WM water partition [0.82ml/g].", argstr="-wmL %f"
    )
    wm_ttt = traits.Float(desc="Time to WM [ATT+0ms].", argstr="-wmTTT %f")

    # *** Segmentation options:
    desc = "Filename of the 4D segmentation (in ASL space) for L/T1 \
estimation and PV correction {WM,GM,CSF}."

    seg = File(exists=True, argstr="-seg %s", desc=desc)
    desc = "Use sigmoid to estimate L from T1: L(T1|gmL,wmL) [Off]."
    sig = traits.Bool(desc=desc, argstr="-sig")
    desc = "Simple PV correction (CBF=vg*CBFg + vw*CBFw, with CBFw=f*CBFg) \
[0.25]."

    pv0 = traits.Int(desc=desc, argstr="-pv0 %d")
    pv2 = traits.Int(desc="In plane PV kernel size [3x3].", argstr="-pv2 %d")
    pv3 = traits.Tuple(
        traits.Int,
        traits.Int,
        traits.Int,
        desc="3D kernel size [3x3x1].",
        argstr="-pv3 %d %d %d",
    )
    desc = "Multiply CBF by this value (e.g. if CL are mislabelled use -1.0)."
    mul = traits.Float(desc=desc, argstr="-mul %f")
    mulgm = traits.Bool(desc="Multiply CBF by segmentation [Off].", argstr="-sig")
    desc = "Set PV threshold for switching off LSQR [O.05]."
    pv_threshold = traits.Bool(desc=desc, argstr="-pvthreshold")
    segstyle = traits.Bool(desc="Set CBF as [gm,wm] not [wm,gm].", argstr="-segstyle")


class FitAslOutputSpec(TraitedSpec):
    """Output Spec for FitAsl."""

    desc = "Filename of the Cerebral Blood Flow map (in ml/100g/min)."
    cbf_file = File(exists=True, desc=desc)
    desc = "Filename of the CBF error map."
    error_file = File(exists=True, desc=desc)
    desc = "Filename of the synthetic ASL data."
    syn_file = File(exists=True, desc=desc)


class FitAsl(NiftyFitCommand):
    """Interface for executable fit_asl from Niftyfit platform.

    Use NiftyFit to perform ASL fitting.

    ASL fitting routines (following EU Cost Action White Paper recommendations)
    Fits Cerebral Blood Flow maps in the first instance.

    `Source code <https://cmiclab.cs.ucl.ac.uk/CMIC/NiftyFit-Release>`_

    Examples
    --------
    >>> from nipype.interfaces import niftyfit
    >>> node = niftyfit.FitAsl()
    >>> node.inputs.source_file = 'asl.nii.gz'
    >>> node.cmdline
    'fit_asl -source asl.nii.gz -cbf asl_cbf.nii.gz -error asl_error.nii.gz \
-syn asl_syn.nii.gz'

    """

    _cmd = get_custom_path("fit_asl", env_dir="NIFTYFITDIR")
    input_spec = FitAslInputSpec
    output_spec = FitAslOutputSpec
    _suffix = "_fit_asl"



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..surface import ModelMaker


def test_ModelMaker_inputs():
    input_map = dict(
        InputVolume=dict(
            argstr="%s",
            extensions=None,
            position=-1,
        ),
        args=dict(
            argstr="%s",
        ),
        color=dict(
            argstr="--color %s",
            extensions=None,
        ),
        debug=dict(
            argstr="--debug ",
        ),
        decimate=dict(
            argstr="--decimate %f",
        ),
        end=dict(
            argstr="--end %d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        filtertype=dict(
            argstr="--filtertype %s",
        ),
        generateAll=dict(
            argstr="--generateAll ",
        ),
        jointsmooth=dict(
            argstr="--jointsmooth ",
        ),
        labels=dict(
            argstr="--labels %s",
            sep=",",
        ),
        modelSceneFile=dict(
            argstr="--modelSceneFile %s...",
            hash_files=False,
        ),
        name=dict(
            argstr="--name %s",
        ),
        pad=dict(
            argstr="--pad ",
        ),
        pointnormals=dict(
            argstr="--pointnormals ",
        ),
        saveIntermediateModels=dict(
            argstr="--saveIntermediateModels ",
        ),
        skipUnNamed=dict(
            argstr="--skipUnNamed ",
        ),
        smooth=dict(
            argstr="--smooth %d",
        ),
        splitnormals=dict(
            argstr="--splitnormals ",
        ),
        start=dict(
            argstr="--start %d",
        ),
    )
    inputs = ModelMaker.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ModelMaker_outputs():
    output_map = dict(
        modelSceneFile=dict(),
    )
    outputs = ModelMaker.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
import os

import nibabel as nb
import numpy as np

from ...utils.filemanip import split_filename, fname_presuffix

from .base import NipyBaseInterface, have_nipy
from ..base import (
    TraitedSpec,
    traits,
    BaseInterfaceInputSpec,
    isdefined,
    File,
    InputMultiPath,
    OutputMultiPath,
)


class ComputeMaskInputSpec(BaseInterfaceInputSpec):
    mean_volume = File(
        exists=True,
        mandatory=True,
        desc="mean EPI image, used to compute the threshold for the mask",
    )
    reference_volume = File(
        exists=True,
        desc=(
            "reference volume used to compute the mask. "
            "If none is give, the mean volume is used."
        ),
    )
    m = traits.Float(desc="lower fraction of the histogram to be discarded")
    M = traits.Float(desc="upper fraction of the histogram to be discarded")
    cc = traits.Bool(desc="Keep only the largest connected component")


class ComputeMaskOutputSpec(TraitedSpec):
    brain_mask = File(exists=True)


class ComputeMask(NipyBaseInterface):
    input_spec = ComputeMaskInputSpec
    output_spec = ComputeMaskOutputSpec

    def _run_interface(self, runtime):
        from nipy.labs.mask import compute_mask

        args = {}
        for key in [
            k
            for k, _ in list(self.inputs.items())
            if k not in BaseInterfaceInputSpec().trait_names()
        ]:
            value = getattr(self.inputs, key)
            if isdefined(value):
                if key in ["mean_volume", "reference_volume"]:
                    value = np.asanyarray(nb.load(value).dataobj)
                args[key] = value

        brain_mask = compute_mask(**args)
        _, name, ext = split_filename(self.inputs.mean_volume)
        self._brain_mask_path = os.path.abspath("%s_mask.%s" % (name, ext))
        nb.save(
            nb.Nifti1Image(brain_mask.astype(np.uint8), nii.affine),
            self._brain_mask_path,
        )

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["brain_mask"] = self._brain_mask_path
        return outputs


class SpaceTimeRealignerInputSpec(BaseInterfaceInputSpec):
    in_file = InputMultiPath(
        File(exists=True), mandatory=True, min_ver="0.4.0.dev", desc="File to realign"
    )
    tr = traits.Float(desc="TR in seconds", requires=["slice_times"])
    slice_times = traits.Either(
        traits.List(traits.Float()),
        traits.Enum(
            "asc_alt_2",
            "asc_alt_2_1",
            "asc_alt_half",
            "asc_alt_siemens",
            "ascending",
            "desc_alt_2",
            "desc_alt_half",
            "descending",
        ),
        desc=("Actual slice acquisition times."),
    )
    slice_info = traits.Either(
        traits.Int,
        traits.List(min_len=2, max_len=2),
        desc=(
            "Single integer or length 2 sequence "
            "If int, the axis in `images` that is the "
            "slice axis.  In a 4D image, this will "
            "often be axis = 2.  If a 2 sequence, then"
            " elements are ``(slice_axis, "
            "slice_direction)``, where ``slice_axis`` "
            "is the slice axis in the image as above, "
            "and ``slice_direction`` is 1 if the "
            "slices were acquired slice 0 first, slice"
            " -1 last, or -1 if acquired slice -1 "
            "first, slice 0 last.  If `slice_info` is "
            "an int, assume "
            "``slice_direction`` == 1."
        ),
        requires=["slice_times"],
    )


class SpaceTimeRealignerOutputSpec(TraitedSpec):
    out_file = OutputMultiPath(File(exists=True), desc="Realigned files")
    par_file = OutputMultiPath(
        File(exists=True),
        desc=("Motion parameter files. Angles are not " "euler angles"),
    )


class SpaceTimeRealigner(NipyBaseInterface):
    """Simultaneous motion and slice timing correction algorithm

    If slice_times is not specified, this algorithm performs spatial motion
    correction

    This interface wraps nipy's SpaceTimeRealign algorithm [Roche2011]_ or simply the
    SpatialRealign algorithm when timing info is not provided.

    Examples
    --------
    >>> from nipype.interfaces.nipy import SpaceTimeRealigner
    >>> #Run spatial realignment only
    >>> realigner = SpaceTimeRealigner()
    >>> realigner.inputs.in_file = ['functional.nii']
    >>> res = realigner.run() # doctest: +SKIP

    >>> realigner = SpaceTimeRealigner()
    >>> realigner.inputs.in_file = ['functional.nii']
    >>> realigner.inputs.tr = 2
    >>> realigner.inputs.slice_times = list(range(0, 3, 67))
    >>> realigner.inputs.slice_info = 2
    >>> res = realigner.run() # doctest: +SKIP


    References
    ----------
    .. [Roche2011] Roche A. A four-dimensional registration algorithm with \
       application to joint correction of motion and slice timing \
       in fMRI. IEEE Trans Med Imaging. 2011 Aug;30(8):1546-54. DOI_.

    .. _DOI: https://doi.org/10.1109/TMI.2011.2131152

    """

    input_spec = SpaceTimeRealignerInputSpec
    output_spec = SpaceTimeRealignerOutputSpec
    keywords = ["slice timing", "motion correction"]

    def _run_interface(self, runtime):
        from nipy import save_image, load_image

        all_ims = [load_image(fname) for fname in self.inputs.in_file]

        if not isdefined(self.inputs.slice_times):
            from nipy.algorithms.registration.groupwise_registration import SpaceRealign

            R = SpaceRealign(all_ims)
        else:
            from nipy.algorithms.registration import SpaceTimeRealign

            R = SpaceTimeRealign(
                all_ims,
                tr=self.inputs.tr,
                slice_times=self.inputs.slice_times,
                slice_info=self.inputs.slice_info,
            )

        R.estimate(refscan=None)

        corr_run = R.resample()
        self._out_file_path = []
        self._par_file_path = []

        for j, corr in enumerate(corr_run):
            self._out_file_path.append(
                os.path.abspath(
                    "corr_%s.nii.gz" % (split_filename(self.inputs.in_file[j])[1])
                )
            )
            save_image(corr, self._out_file_path[j])

            self._par_file_path.append(
                os.path.abspath("%s.par" % (os.path.split(self.inputs.in_file[j])[1]))
            )
            mfile = open(self._par_file_path[j], "w")
            motion = R._transforms[j]
            # nipy does not encode euler angles. return in original form of
            # translation followed by rotation vector see:
            # http://en.wikipedia.org/wiki/Rodrigues'_rotation_formula
            for i, mo in enumerate(motion):
                params = [
                    "%.10f" % item for item in np.hstack((mo.translation, mo.rotation))
                ]
                string = " ".join(params) + "\n"
                mfile.write(string)
            mfile.close()

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = self._out_file_path
        outputs["par_file"] = self._par_file_path
        return outputs


class TrimInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="EPI image to trim")
    begin_index = traits.Int(0, usedefault=True, desc="first volume")
    end_index = traits.Int(
        0, usedefault=True, desc="last volume indexed as in python (and 0 for last)"
    )
    out_file = File(desc="output filename")
    suffix = traits.Str(
        "_trim",
        usedefault=True,
        desc="suffix for out_file to use if no out_file provided",
    )


class TrimOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class Trim(NipyBaseInterface):
    """Simple interface to trim a few volumes from a 4d fmri nifti file

    Examples
    --------
    >>> from nipype.interfaces.nipy.preprocess import Trim
    >>> trim = Trim()
    >>> trim.inputs.in_file = 'functional.nii'
    >>> trim.inputs.begin_index = 3 # remove 3 first volumes
    >>> res = trim.run() # doctest: +SKIP

    """

    input_spec = TrimInputSpec
    output_spec = TrimOutputSpec

    def _run_interface(self, runtime):
        out_file = self._list_outputs()["out_file"]
        nii = nb.load(self.inputs.in_file)
        if self.inputs.end_index == 0:
            s = slice(self.inputs.begin_index, nii.shape[3])
        else:
            s = slice(self.inputs.begin_index, self.inputs.end_index)
        nii2 = nb.Nifti1Image(nii.dataobj[..., s], nii.affine, nii.header)
        nb.save(nii2, out_file)
        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_file
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = fname_presuffix(
                self.inputs.in_file, newpath=os.getcwd(), suffix=self.inputs.suffix
            )
        outputs["out_file"] = os.path.abspath(outputs["out_file"])
        return outputs



# -*- coding: utf-8 -*-
"""This script provides interfaces for BrainSuite command line tools.
Please see brainsuite.org for more information.

Author: Jason Wong
"""

import os
import re as regex

from ..base import (
    TraitedSpec,
    CommandLineInputSpec,
    CommandLine,
    File,
    traits,
    isdefined,
)


class BseInputSpec(CommandLineInputSpec):

    inputMRIFile = File(mandatory=True, argstr="-i %s", desc="input MRI volume")
    outputMRIVolume = File(
        desc="output brain-masked MRI volume. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        hash_files=False,
        genfile=True,
    )
    outputMaskFile = File(
        desc="save smooth brain mask. If unspecified, output file name will be auto generated.",
        argstr="--mask %s",
        hash_files=False,
        genfile=True,
    )
    diffusionConstant = traits.Float(
        25, usedefault=True, desc="diffusion constant", argstr="-d %f"
    )
    diffusionIterations = traits.Int(
        3, usedefault=True, desc="diffusion iterations", argstr="-n %d"
    )
    edgeDetectionConstant = traits.Float(
        0.64, usedefault=True, desc="edge detection constant", argstr="-s %f"
    )
    radius = traits.Float(
        1, usedefault=True, desc="radius of erosion/dilation filter", argstr="-r %f"
    )
    dilateFinalMask = traits.Bool(
        True, usedefault=True, desc="dilate final mask", argstr="-p"
    )
    trim = traits.Bool(True, usedefault=True, desc="trim brainstem", argstr="--trim")
    outputDiffusionFilter = File(
        desc="diffusion filter output", argstr="--adf %s", hash_files=False
    )
    outputEdgeMap = File(desc="edge map output", argstr="--edge %s", hash_files=False)
    outputDetailedBrainMask = File(
        desc="save detailed brain mask", argstr="--hires %s", hash_files=False
    )
    outputCortexFile = File(desc="cortex file", argstr="--cortex %s", hash_files=False)
    verbosityLevel = traits.Float(
        1, usedefault=True, desc=" verbosity level (0=silent)", argstr="-v %f"
    )
    noRotate = traits.Bool(
        desc="retain original orientation(default behavior will auto-rotate input NII files to LPI orientation)",
        argstr="--norotate",
    )
    timer = traits.Bool(desc="show timing", argstr="--timer")


class BseOutputSpec(TraitedSpec):
    outputMRIVolume = File(desc="path/name of brain-masked MRI volume")
    outputMaskFile = File(desc="path/name of smooth brain mask")
    outputDiffusionFilter = File(desc="path/name of diffusion filter output")
    outputEdgeMap = File(desc="path/name of edge map output")
    outputDetailedBrainMask = File(desc="path/name of detailed brain mask")
    outputCortexFile = File(desc="path/name of cortex file")


class Bse(CommandLine):
    """
    brain surface extractor (BSE)
    This program performs automated skull and scalp removal on T1-weighted MRI volumes.

    http://brainsuite.org/processing/surfaceextraction/bse/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> bse = brainsuite.Bse()
    >>> bse.inputs.inputMRIFile = example_data('structural.nii')
    >>> results = bse.run() #doctest: +SKIP

    """

    input_spec = BseInputSpec
    output_spec = BseOutputSpec
    _cmd = "bse"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        fileToSuffixMap = {
            "outputMRIVolume": ".bse.nii.gz",
            "outputMaskFile": ".mask.nii.gz",
        }

        if name in fileToSuffixMap:
            return getFileName(self.inputs.inputMRIFile, fileToSuffixMap[name])

        return None

    def _list_outputs(self):
        return l_outputs(self)


class BfcInputSpec(CommandLineInputSpec):
    inputMRIFile = File(
        mandatory=True, desc="input skull-stripped MRI volume", argstr="-i %s"
    )
    inputMaskFile = File(desc="mask file", argstr="-m %s", hash_files=False)
    outputMRIVolume = File(
        desc="output bias-corrected MRI volume. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        hash_files=False,
        genfile=True,
    )
    outputBiasField = File(
        desc="save bias field estimate", argstr="--bias %s", hash_files=False
    )
    outputMaskedBiasField = File(
        desc="save bias field estimate (masked)",
        argstr="--maskedbias %s",
        hash_files=False,
    )
    histogramRadius = traits.Int(desc="histogram radius (voxels)", argstr="-r %d")
    biasEstimateSpacing = traits.Int(
        desc="bias sample spacing (voxels)", argstr="-s %d"
    )
    controlPointSpacing = traits.Int(
        desc="control point spacing (voxels)", argstr="-c %d"
    )
    splineLambda = traits.Float(
        desc="spline stiffness weighting parameter", argstr="-w %f"
    )
    histogramType = traits.Enum(
        "ellipse",
        "block",
        desc="""\
Options for type of histogram:

  * ``ellipse``: use ellipsoid for ROI histogram
  * ``block``:use block for ROI histogram

""",
        argstr="%s",
    )
    iterativeMode = traits.Bool(
        desc="iterative mode (overrides -r, -s, -c, -w settings)", argstr="--iterate"
    )
    correctionScheduleFile = File(desc="list of parameters ", argstr="--schedule %s")
    biasFieldEstimatesOutputPrefix = traits.Str(
        desc="save iterative bias field estimates as <prefix>.n.field.nii.gz",
        argstr="--biasprefix %s",
    )
    correctedImagesOutputPrefix = traits.Str(
        desc="save iterative corrected images as <prefix>.n.bfc.nii.gz",
        argstr="--prefix %s",
    )
    correctWholeVolume = traits.Bool(
        desc="apply correction field to entire volume", argstr="--extrapolate"
    )
    minBias = traits.Float(
        0.5, usedefault=True, desc="minimum allowed bias value", argstr="-L %f"
    )
    maxBias = traits.Float(
        1.5, usedefault=True, desc="maximum allowed bias value", argstr="-U %f"
    )
    biasRange = traits.Enum(
        "low",
        "medium",
        "high",
        desc="""\
Preset options for bias_model

  * low: small bias model [0.95,1.05]
  * medium: medium bias model [0.90,1.10]
  * high: high bias model [0.80,1.20]

""",
        argstr="%s",
    )
    intermediate_file_type = traits.Enum(
        "analyze",
        "nifti",
        "gzippedAnalyze",
        "gzippedNifti",
        desc="Options for the format in which intermediate files are generated",
        argstr="%s",
    )
    convergenceThreshold = traits.Float(desc="convergence threshold", argstr="--eps %f")
    biasEstimateConvergenceThreshold = traits.Float(
        desc="bias estimate convergence threshold (values > 0.1 disable)",
        argstr="--beps %f",
    )
    verbosityLevel = traits.Int(desc="verbosity level (0=silent)", argstr="-v %d")
    timer = traits.Bool(desc="display timing information", argstr="--timer")


class BfcOutputSpec(TraitedSpec):
    outputMRIVolume = File(desc="path/name of output file")
    outputBiasField = File(desc="path/name of bias field output file")
    outputMaskedBiasField = File(desc="path/name of masked bias field output")
    correctionScheduleFile = File(desc="path/name of schedule file")


class Bfc(CommandLine):
    """
    bias field corrector (BFC)
    This program corrects gain variation in T1-weighted MRI.

    http://brainsuite.org/processing/surfaceextraction/bfc/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> bfc = brainsuite.Bfc()
    >>> bfc.inputs.inputMRIFile = example_data('structural.nii')
    >>> bfc.inputs.inputMaskFile = example_data('mask.nii')
    >>> results = bfc.run() #doctest: +SKIP

    """

    input_spec = BfcInputSpec
    output_spec = BfcOutputSpec
    _cmd = "bfc"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        fileToSuffixMap = {"outputMRIVolume": ".bfc.nii.gz"}
        if name in fileToSuffixMap:
            return getFileName(self.inputs.inputMRIFile, fileToSuffixMap[name])

        return None

    def _format_arg(self, name, spec, value):
        if name == "histogramType":
            return spec.argstr % {"ellipse": "--ellipse", "block": "--block"}[value]
        if name == "biasRange":
            return (
                spec.argstr
                % {"low": "--low", "medium": "--medium", "high": "--high"}[value]
            )
        if name == "intermediate_file_type":
            return spec.argstr % {
                "analyze": "--analyze",
                "nifti": "--nifti",
                "gzippedAnalyze": "--analyzegz",
                "gzippedNifti": "--niftigz",
            }[value]

        return super(Bfc, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        return l_outputs(self)


class PvcInputSpec(CommandLineInputSpec):
    inputMRIFile = File(mandatory=True, desc="MRI file", argstr="-i %s")
    inputMaskFile = File(desc="brain mask file", argstr="-m %s")
    outputLabelFile = File(
        desc="output label file. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    outputTissueFractionFile = File(
        desc="output tissue fraction file", argstr="-f %s", genfile=True
    )
    spatialPrior = traits.Float(desc="spatial prior strength", argstr="-l %f")
    verbosity = traits.Int(desc="verbosity level (0 = silent)", argstr="-v %d")
    threeClassFlag = traits.Bool(
        desc="use a three-class (CSF=0,GM=1,WM=2) labeling", argstr="-3"
    )
    timer = traits.Bool(desc="time processing", argstr="--timer")


class PvcOutputSpec(TraitedSpec):
    outputLabelFile = File(desc="path/name of label file")
    outputTissueFractionFile = File(desc="path/name of tissue fraction file")


class Pvc(CommandLine):
    """
    partial volume classifier (PVC) tool.
    This program performs voxel-wise tissue classification T1-weighted MRI.
    Image should be skull-stripped and bias-corrected before tissue classification.

    http://brainsuite.org/processing/surfaceextraction/pvc/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> pvc = brainsuite.Pvc()
    >>> pvc.inputs.inputMRIFile = example_data('structural.nii')
    >>> pvc.inputs.inputMaskFile = example_data('mask.nii')
    >>> results = pvc.run() #doctest: +SKIP

    """

    input_spec = PvcInputSpec
    output_spec = PvcOutputSpec
    _cmd = "pvc"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        fileToSuffixMap = {
            "outputLabelFile": ".pvc.label.nii.gz",
            "outputTissueFractionFile": ".pvc.frac.nii.gz",
        }
        if name in fileToSuffixMap:
            return getFileName(self.inputs.inputMRIFile, fileToSuffixMap[name])

        return None

    def _list_outputs(self):
        return l_outputs(self)


class CerebroInputSpec(CommandLineInputSpec):
    inputMRIFile = File(mandatory=True, desc="input 3D MRI volume", argstr="-i %s")
    inputAtlasMRIFile = File(
        mandatory=True, desc="atlas MRI volume", argstr="--atlas %s"
    )
    inputAtlasLabelFile = File(
        mandatory=True, desc="atlas labeling", argstr="--atlaslabels %s"
    )
    inputBrainMaskFile = File(desc="brain mask file", argstr="-m %s")
    outputCerebrumMaskFile = File(
        desc="output cerebrum mask volume. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    outputLabelVolumeFile = File(
        desc="output labeled hemisphere/cerebrum volume. If unspecified, output file name will be auto generated.",
        argstr="-l %s",
        genfile=True,
    )
    costFunction = traits.Int(2, usedefault=True, desc="0,1,2", argstr="-c %d")
    useCentroids = traits.Bool(
        desc="use centroids of data to initialize position", argstr="--centroids"
    )
    outputAffineTransformFile = File(
        desc="save affine transform to file.", argstr="--air %s", genfile=True
    )
    outputWarpTransformFile = File(
        desc="save warp transform to file.", argstr="--warp %s", genfile=True
    )
    verbosity = traits.Int(desc="verbosity level (0=silent)", argstr="-v %d")
    linearConvergence = traits.Float(desc="linear convergence", argstr="--linconv %f")
    warpLabel = traits.Int(desc="warp order (2,3,4,5,6,7,8)", argstr="--warplevel %d")
    warpConvergence = traits.Float(desc="warp convergence", argstr="--warpconv %f")
    keepTempFiles = traits.Bool(desc="don't remove temporary files", argstr="--keep")
    tempDirectory = traits.Str(
        desc="specify directory to use for temporary files", argstr="--tempdir %s"
    )
    tempDirectoryBase = traits.Str(
        desc="create a temporary directory within this directory",
        argstr="--tempdirbase %s",
    )


class CerebroOutputSpec(TraitedSpec):
    outputCerebrumMaskFile = File(desc="path/name of cerebrum mask file")
    outputLabelVolumeFile = File(desc="path/name of label mask file")
    outputAffineTransformFile = File(desc="path/name of affine transform file")
    outputWarpTransformFile = File(desc="path/name of warp transform file")


class Cerebro(CommandLine):
    """
    Cerebrum/cerebellum labeling tool
    This program performs automated labeling of cerebellum and cerebrum in T1 MRI.
    Input MRI should be skull-stripped or a brain-only mask should be provided.


    http://brainsuite.org/processing/surfaceextraction/cerebrum/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> cerebro = brainsuite.Cerebro()
    >>> cerebro.inputs.inputMRIFile = example_data('structural.nii')
    >>> cerebro.inputs.inputAtlasMRIFile = 'atlasMRIVolume.img'
    >>> cerebro.inputs.inputAtlasLabelFile = 'atlasLabels.img'
    >>> cerebro.inputs.inputBrainMaskFile = example_data('mask.nii')
    >>> results = cerebro.run() #doctest: +SKIP

    """

    input_spec = CerebroInputSpec
    output_spec = CerebroOutputSpec
    _cmd = "cerebro"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        fileToSuffixMap = {
            "outputCerebrumMaskFile": ".cerebrum.mask.nii.gz",
            "outputLabelVolumeFile": ".hemi.label.nii.gz",
            "outputWarpTransformFile": ".warp",
            "outputAffineTransformFile": ".air",
        }
        if name in fileToSuffixMap:
            return getFileName(self.inputs.inputMRIFile, fileToSuffixMap[name])

        return None

    def _list_outputs(self):
        return l_outputs(self)


class CortexInputSpec(CommandLineInputSpec):
    inputHemisphereLabelFile = File(
        mandatory=True, desc="hemisphere / lobe label volume", argstr="-h %s"
    )
    outputCerebrumMask = File(
        desc="output structure mask. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    inputTissueFractionFile = File(
        mandatory=True, desc="tissue fraction file (32-bit float)", argstr="-f %s"
    )
    tissueFractionThreshold = traits.Float(
        50.0,
        usedefault=True,
        desc="tissue fraction threshold (percentage)",
        argstr="-p %f",
    )
    computeWGBoundary = traits.Bool(
        True, usedefault=True, desc="compute WM/GM boundary", argstr="-w"
    )
    computeGCBoundary = traits.Bool(desc="compute GM/CSF boundary", argstr="-g")
    includeAllSubcorticalAreas = traits.Bool(
        True,
        usedefault=True,
        desc="include all subcortical areas in WM mask",
        argstr="-a",
    )
    verbosity = traits.Int(desc="verbosity level", argstr="-v %d")
    timer = traits.Bool(desc="timing function", argstr="--timer")


class CortexOutputSpec(TraitedSpec):
    outputCerebrumMask = File(desc="path/name of cerebrum mask")


class Cortex(CommandLine):
    """
    cortex extractor
    This program produces a cortical mask using tissue fraction estimates
    and a co-registered cerebellum/hemisphere mask.

    http://brainsuite.org/processing/surfaceextraction/cortex/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> cortex = brainsuite.Cortex()
    >>> cortex.inputs.inputHemisphereLabelFile = example_data('mask.nii')
    >>> cortex.inputs.inputTissueFractionFile = example_data('tissues.nii.gz')
    >>> results = cortex.run() #doctest: +SKIP

    """

    input_spec = CortexInputSpec
    output_spec = CortexOutputSpec
    _cmd = "cortex"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        if name == "outputCerebrumMask":
            return getFileName(
                self.inputs.inputHemisphereLabelFile, ".init.cortex.mask.nii.gz"
            )
        return None

    def _list_outputs(self):
        return l_outputs(self)


class ScrubmaskInputSpec(CommandLineInputSpec):
    inputMaskFile = File(
        mandatory=True, desc="input structure mask file", argstr="-i %s"
    )
    outputMaskFile = File(
        desc="output structure mask file. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    backgroundFillThreshold = traits.Int(
        2, usedefault=True, desc="background fill threshold", argstr="-b %d"
    )
    foregroundTrimThreshold = traits.Int(
        0, usedefault=True, desc="foreground trim threshold", argstr="-f %d"
    )
    numberIterations = traits.Int(desc="number of iterations", argstr="-n %d")
    verbosity = traits.Int(desc="verbosity (0=silent)", argstr="-v %d")
    timer = traits.Bool(desc="timing function", argstr="--timer")


class ScrubmaskOutputSpec(TraitedSpec):
    outputMaskFile = File(desc="path/name of mask file")


class Scrubmask(CommandLine):
    """
    ScrubMask tool
    scrubmask filters binary masks to trim loosely connected voxels that may
    result from segmentation errors and produce bumps on tessellated surfaces.

    http://brainsuite.org/processing/surfaceextraction/scrubmask/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> scrubmask = brainsuite.Scrubmask()
    >>> scrubmask.inputs.inputMaskFile = example_data('mask.nii')
    >>> results = scrubmask.run() #doctest: +SKIP

    """

    input_spec = ScrubmaskInputSpec
    output_spec = ScrubmaskOutputSpec
    _cmd = "scrubmask"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        if name == "outputMaskFile":
            return getFileName(
                self.inputs.inputMaskFile, ".cortex.scrubbed.mask.nii.gz"
            )

        return None

    def _list_outputs(self):
        return l_outputs(self)


class TcaInputSpec(CommandLineInputSpec):
    inputMaskFile = File(mandatory=True, desc="input mask volume", argstr="-i %s")
    outputMaskFile = File(
        desc="output mask volume. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    minCorrectionSize = traits.Int(
        2500, usedefault=True, desc="maximum correction size", argstr="-m %d"
    )
    maxCorrectionSize = traits.Int(desc="minimum correction size", argstr="-n %d")
    foregroundDelta = traits.Int(
        20, usedefault=True, desc="foreground delta", argstr="--delta %d"
    )
    verbosity = traits.Int(desc="verbosity (0 = quiet)", argstr="-v %d")
    timer = traits.Bool(desc="timing function", argstr="--timer")


class TcaOutputSpec(TraitedSpec):
    outputMaskFile = File(desc="path/name of mask file")


class Tca(CommandLine):
    """
    topological correction algorithm (TCA)
    This program removes topological handles from a binary object.

    http://brainsuite.org/processing/surfaceextraction/tca/

    Examples
    --------
    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> tca = brainsuite.Tca()
    >>> tca.inputs.inputMaskFile = example_data('mask.nii')
    >>> results = tca.run() #doctest: +SKIP

    """

    input_spec = TcaInputSpec
    output_spec = TcaOutputSpec
    _cmd = "tca"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        if name == "outputMaskFile":
            return getFileName(self.inputs.inputMaskFile, ".cortex.tca.mask.nii.gz")

        return None

    def _list_outputs(self):
        return l_outputs(self)


class DewispInputSpec(CommandLineInputSpec):
    inputMaskFile = File(mandatory=True, desc="input file", argstr="-i %s")
    outputMaskFile = File(
        desc="output file. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    verbosity = traits.Int(desc="verbosity", argstr="-v %d")
    sizeThreshold = traits.Int(desc="size threshold", argstr="-t %d")
    maximumIterations = traits.Int(desc="maximum number of iterations", argstr="-n %d")
    timer = traits.Bool(desc="time processing", argstr="--timer")


class DewispOutputSpec(TraitedSpec):
    outputMaskFile = File(desc="path/name of mask file")


class Dewisp(CommandLine):
    """
    dewisp
    removes wispy tendril structures from cortex model binary masks.
    It does so based on graph theoretic analysis of connected components,
    similar to TCA. Each branch of the structure graph is analyzed to determine
    pinch points that indicate a likely error in segmentation that attaches noise
    to the image. The pinch threshold determines how many voxels the cross-section
    can be before it is considered part of the image.

    http://brainsuite.org/processing/surfaceextraction/dewisp/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> dewisp = brainsuite.Dewisp()
    >>> dewisp.inputs.inputMaskFile = example_data('mask.nii')
    >>> results = dewisp.run() #doctest: +SKIP

    """

    input_spec = DewispInputSpec
    output_spec = DewispOutputSpec
    _cmd = "dewisp"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        if name == "outputMaskFile":
            return getFileName(self.inputs.inputMaskFile, ".cortex.dewisp.mask.nii.gz")

        return None

    def _list_outputs(self):
        return l_outputs(self)


class DfsInputSpec(CommandLineInputSpec):
    inputVolumeFile = File(mandatory=True, desc="input 3D volume", argstr="-i %s")
    outputSurfaceFile = File(
        desc="output surface mesh file. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    inputShadingVolume = File(
        desc="shade surface model with data from image volume", argstr="-c %s"
    )
    smoothingIterations = traits.Int(
        10, usedefault=True, desc="number of smoothing iterations", argstr="-n %d"
    )
    smoothingConstant = traits.Float(
        0.5, usedefault=True, desc="smoothing constant", argstr="-a %f"
    )
    curvatureWeighting = traits.Float(
        5.0, usedefault=True, desc="curvature weighting", argstr="-w %f"
    )
    scalingPercentile = traits.Float(desc="scaling percentile", argstr="-f %f")
    nonZeroTessellation = traits.Bool(
        desc="tessellate non-zero voxels",
        argstr="-nz",
        xor=("nonZeroTessellation", "specialTessellation"),
    )
    tessellationThreshold = traits.Float(
        desc="To be used with specialTessellation. Set this value first, then set specialTessellation value.\nUsage: tessellate voxels greater_than, less_than, or equal_to <tessellationThreshold>",
        argstr="%f",
    )
    specialTessellation = traits.Enum(
        "greater_than",
        "less_than",
        "equal_to",
        desc="To avoid throwing a UserWarning, set tessellationThreshold first. Then set this attribute.\nUsage: tessellate voxels greater_than, less_than, or equal_to <tessellationThreshold>",
        argstr="%s",
        xor=("nonZeroTessellation", "specialTessellation"),
        requires=["tessellationThreshold"],
        position=-1,
    )
    zeroPadFlag = traits.Bool(
        desc="zero-pad volume (avoids clipping at edges)", argstr="-z"
    )
    noNormalsFlag = traits.Bool(
        desc="do not compute vertex normals", argstr="--nonormals"
    )
    postSmoothFlag = traits.Bool(
        desc="smooth vertices after coloring", argstr="--postsmooth"
    )
    verbosity = traits.Int(desc="verbosity (0 = quiet)", argstr="-v %d")
    timer = traits.Bool(desc="timing function", argstr="--timer")


class DfsOutputSpec(TraitedSpec):
    outputSurfaceFile = File(desc="path/name of surface file")


class Dfs(CommandLine):
    """
    Surface Generator
    Generates mesh surfaces using an isosurface algorithm.

    http://brainsuite.org/processing/surfaceextraction/inner-cortical-surface/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> dfs = brainsuite.Dfs()
    >>> dfs.inputs.inputVolumeFile = example_data('structural.nii')
    >>> results = dfs.run() #doctest: +SKIP

    """

    input_spec = DfsInputSpec
    output_spec = DfsOutputSpec
    _cmd = "dfs"

    def _format_arg(self, name, spec, value):
        if name == "tessellationThreshold":
            return ""  # blank argstr
        if name == "specialTessellation":
            threshold = self.inputs.tessellationThreshold
            return spec.argstr % {
                "greater_than": "".join(("-gt %f" % threshold)),
                "less_than": "".join(("-lt %f" % threshold)),
                "equal_to": "".join(("-eq %f" % threshold)),
            }[value]
        return super(Dfs, self)._format_arg(name, spec, value)

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        if name == "outputSurfaceFile":
            return getFileName(self.inputs.inputVolumeFile, ".inner.cortex.dfs")

        return None

    def _list_outputs(self):
        return l_outputs(self)


class PialmeshInputSpec(CommandLineInputSpec):
    inputSurfaceFile = File(mandatory=True, desc="input file", argstr="-i %s")
    outputSurfaceFile = File(
        desc="output file. If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    verbosity = traits.Int(desc="verbosity", argstr="-v %d")
    inputTissueFractionFile = File(
        mandatory=True, desc="floating point (32) tissue fraction image", argstr="-f %s"
    )
    numIterations = traits.Int(
        100, usedefault=True, desc="number of iterations", argstr="-n %d"
    )
    searchRadius = traits.Float(
        1, usedefault=True, desc="search radius", argstr="-r %f"
    )
    stepSize = traits.Float(0.4, usedefault=True, desc="step size", argstr="-s %f")
    inputMaskFile = File(
        mandatory=True, desc="restrict growth to mask file region", argstr="-m %s"
    )
    maxThickness = traits.Float(
        20, usedefault=True, desc="maximum allowed tissue thickness", argstr="--max %f"
    )
    tissueThreshold = traits.Float(
        1.05, usedefault=True, desc="tissue threshold", argstr="-t %f"
    )
    # output interval is not an output -- it specifies how frequently the
    # output surfaces are generated
    outputInterval = traits.Int(
        10, usedefault=True, desc="output interval", argstr="--interval %d"
    )
    exportPrefix = traits.Str(
        desc="prefix for exporting surfaces if interval is set", argstr="--prefix %s"
    )
    laplacianSmoothing = traits.Float(
        0.025, usedefault=True, desc="apply Laplacian smoothing", argstr="--smooth %f"
    )
    timer = traits.Bool(desc="show timing", argstr="--timer")
    recomputeNormals = traits.Bool(
        desc="recompute normals at each iteration", argstr="--norm"
    )
    normalSmoother = traits.Float(
        0.2, usedefault=True, desc="strength of normal smoother.", argstr="--nc %f"
    )
    tangentSmoother = traits.Float(
        desc="strength of tangential smoother.", argstr="--tc %f"
    )


class PialmeshOutputSpec(TraitedSpec):
    outputSurfaceFile = File(desc="path/name of surface file")


class Pialmesh(CommandLine):
    """
    pialmesh
    computes a pial surface model using an inner WM/GM mesh and a tissue fraction map.

    http://brainsuite.org/processing/surfaceextraction/pial/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> pialmesh = brainsuite.Pialmesh()
    >>> pialmesh.inputs.inputSurfaceFile = 'input_mesh.dfs'
    >>> pialmesh.inputs.inputTissueFractionFile = 'frac_file.nii.gz'
    >>> pialmesh.inputs.inputMaskFile = example_data('mask.nii')
    >>> results = pialmesh.run() #doctest: +SKIP

    """

    input_spec = PialmeshInputSpec
    output_spec = PialmeshOutputSpec
    _cmd = "pialmesh"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        if name == "outputSurfaceFile":
            return getFileName(self.inputs.inputSurfaceFile, ".pial.cortex.dfs")

        return None

    def _list_outputs(self):
        return l_outputs(self)


class HemisplitInputSpec(CommandLineInputSpec):
    inputSurfaceFile = File(mandatory=True, desc="input surface", argstr="-i %s")
    inputHemisphereLabelFile = File(
        mandatory=True, desc="input hemisphere label volume", argstr="-l %s"
    )
    outputLeftHemisphere = File(
        desc="output surface file, left hemisphere. If unspecified, output file name will be auto generated.",
        argstr="--left %s",
        genfile=True,
    )
    outputRightHemisphere = File(
        desc="output surface file, right hemisphere. If unspecified, output file name will be auto generated.",
        argstr="--right %s",
        genfile=True,
    )
    pialSurfaceFile = File(
        desc="pial surface file -- must have same geometry as input surface",
        argstr="-p %s",
    )
    outputLeftPialHemisphere = File(
        desc="output pial surface file, left hemisphere. If unspecified, output file name will be auto generated.",
        argstr="-pl %s",
        genfile=True,
    )
    outputRightPialHemisphere = File(
        desc="output pial surface file, right hemisphere. If unspecified, output file name will be auto generated.",
        argstr="-pr %s",
        genfile=True,
    )
    verbosity = traits.Int(desc="verbosity (0 = silent)", argstr="-v %d")
    timer = traits.Bool(desc="timing function", argstr="--timer")


class HemisplitOutputSpec(TraitedSpec):
    outputLeftHemisphere = File(desc="path/name of left hemisphere")
    outputRightHemisphere = File(desc="path/name of right hemisphere")
    outputLeftPialHemisphere = File(desc="path/name of left pial hemisphere")
    outputRightPialHemisphere = File(desc="path/name of right pial hemisphere")


class Hemisplit(CommandLine):
    """
    Hemisphere splitter
    Splits a surface object into two separate surfaces given an input label volume.
    Each vertex is labeled left or right based on the labels being odd (left) or even (right).
    The largest contour on the split surface is then found and used as the separation between left and right.

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> hemisplit = brainsuite.Hemisplit()
    >>> hemisplit.inputs.inputSurfaceFile = 'input_surf.dfs'
    >>> hemisplit.inputs.inputHemisphereLabelFile = 'label.nii'
    >>> hemisplit.inputs.pialSurfaceFile = 'pial.dfs'
    >>> results = hemisplit.run() #doctest: +SKIP

    """

    input_spec = HemisplitInputSpec
    output_spec = HemisplitOutputSpec
    _cmd = "hemisplit"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        fileToSuffixMap = {
            "outputLeftHemisphere": ".left.inner.cortex.dfs",
            "outputLeftPialHemisphere": ".left.pial.cortex.dfs",
            "outputRightHemisphere": ".right.inner.cortex.dfs",
            "outputRightPialHemisphere": ".right.pial.cortex.dfs",
        }
        if name in fileToSuffixMap:
            return getFileName(self.inputs.inputSurfaceFile, fileToSuffixMap[name])

        return None

    def _list_outputs(self):
        return l_outputs(self)


class SkullfinderInputSpec(CommandLineInputSpec):
    inputMRIFile = File(mandatory=True, desc="input file", argstr="-i %s")
    inputMaskFile = File(
        mandatory=True,
        desc="A brain mask file, 8-bit image (0=non-brain, 255=brain)",
        argstr="-m %s",
    )
    outputLabelFile = File(
        desc="output multi-colored label volume segmenting brain, scalp, inner skull & outer skull "
        "If unspecified, output file name will be auto generated.",
        argstr="-o %s",
        genfile=True,
    )
    verbosity = traits.Int(desc="verbosity", argstr="-v %d")
    lowerThreshold = traits.Int(desc="Lower threshold for segmentation", argstr="-l %d")
    upperThreshold = traits.Int(desc="Upper threshold for segmentation", argstr="-u %d")
    surfaceFilePrefix = traits.Str(
        desc="if specified, generate surface files for brain, skull, and scalp",
        argstr="-s %s",
    )
    bgLabelValue = traits.Int(
        desc="background label value (0-255)", argstr="--bglabel %d"
    )
    scalpLabelValue = traits.Int(
        desc="scalp label value (0-255)", argstr="--scalplabel %d"
    )
    skullLabelValue = traits.Int(
        desc="skull label value (0-255)", argstr="--skulllabel %d"
    )
    spaceLabelValue = traits.Int(
        desc="space label value (0-255)", argstr="--spacelabel %d"
    )
    brainLabelValue = traits.Int(
        desc="brain label value (0-255)", argstr="--brainlabel %d"
    )
    performFinalOpening = traits.Bool(
        desc="perform a final opening operation on the scalp mask",
        argstr="--finalOpening",
    )


class SkullfinderOutputSpec(TraitedSpec):
    outputLabelFile = File(desc="path/name of label file")


class Skullfinder(CommandLine):
    """
    Skull and scalp segmentation algorithm.

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> from nipype.testing import example_data
    >>> skullfinder = brainsuite.Skullfinder()
    >>> skullfinder.inputs.inputMRIFile = example_data('structural.nii')
    >>> skullfinder.inputs.inputMaskFile = example_data('mask.nii')
    >>> results = skullfinder.run() #doctest: +SKIP

    """

    input_spec = SkullfinderInputSpec
    output_spec = SkullfinderOutputSpec
    _cmd = "skullfinder"

    def _gen_filename(self, name):
        inputs = self.inputs.get()
        if isdefined(inputs[name]):
            return os.path.abspath(inputs[name])

        if name == "outputLabelFile":
            return getFileName(self.inputs.inputMRIFile, ".skullfinder.label.nii.gz")

        return None

    def _list_outputs(self):
        return l_outputs(self)


class SVRegInputSpec(CommandLineInputSpec):
    subjectFilePrefix = traits.Str(
        argstr="'%s'",
        mandatory=True,
        position=0,
        desc="Absolute path and filename prefix of the subjects output from BrainSuite "
        "Cortical Surface Extraction Sequence",
    )
    dataSinkDelay = traits.List(
        traits.Str,
        argstr="%s",
        desc="Connect datasink out_file to dataSinkDelay to delay execution of SVReg "
        "until dataSink has finished sinking CSE outputs."
        "For use with parallel processing workflows including Brainsuites Cortical "
        "Surface Extraction sequence (SVReg requires certain files from Brainsuite "
        "CSE, which must all be in the pathway specified by subjectFilePrefix. see "
        "http://brainsuite.org/processing/svreg/usage/ for list of required inputs ",
    )
    atlasFilePrefix = traits.Str(
        position=1,
        argstr="'%s'",
        desc="Optional: Absolute Path and filename prefix of atlas files and labels to which "
        "the subject will be registered. If unspecified, SVReg"
        "will use its own included atlas files",
    )
    iterations = traits.Int(
        argstr="'-H %d'",
        desc="Assigns a number of iterations in the intensity registration step."
        "if unspecified, performs 100 iterations",
    )
    refineOutputs = traits.Bool(
        argstr="'-r'", desc="Refine outputs at the expense of more processing time."
    )
    skipToVolumeReg = traits.Bool(
        argstr="'-s'",
        desc="If surface registration was already performed at an earlier time and the "
        "user would not like to redo this step, then this flag may be used to skip "
        "ahead to the volumetric registration. Necessary input files will need to "
        "be present in the input directory called by the command.",
    )
    skipToIntensityReg = traits.Bool(
        argstr="'-p'",
        desc="If the p-harmonic volumetric registration was already performed at an "
        "earlier time and the user would not like to redo this step, then this "
        "flag may be used to skip ahead to the intensity registration and "
        "label transfer step.",
    )
    useManualMaskFile = traits.Bool(
        argstr="'-cbm'",
        desc="Can call a manually edited cerebrum mask to limit boundaries. Will "
        "use file: subbasename.cerebrum.mask.nii.gz Make sure to correctly "
        "replace your manually edited mask file in your input folder with the "
        "correct subbasename.",
    )
    curveMatchingInstructions = traits.Str(
        argstr="'-cur %s'",
        desc="Used to take control of the curve matching process between the atlas "
        "and subject. One can specify the name of the .dfc file <sulname.dfc> and "
        "the sulcal numbers <#sul> to be used as constraints. "
        'example: curveMatchingInstructions = "subbasename.right.dfc 1 2 20"',
    )
    useCerebrumMask = traits.Bool(
        argstr="'-C'",
        desc="The cerebrum mask <subbasename.cerebrum.mask.nii.gz> will be used for "
        "masking the final labels instead of the default pial surface mask. "
        "Every voxel will be labeled within the cerebrum mask regardless of "
        "the boundaries of the pial surface.",
    )
    pialSurfaceMaskDilation = traits.Int(
        argstr="'-D %d'",
        desc="Cortical volume labels found in file output subbasename.svreg.label.nii.gz "
        "find its boundaries by using the pial surface then dilating by 1 voxel. "
        "Use this flag in order to control the number of pial surface mask dilation. "
        "(ie. -D 0 will assign no voxel dilation)",
    )
    keepIntermediates = traits.Bool(
        argstr="'-k'",
        desc="Keep the intermediate files after the svreg sequence is complete.",
    )
    _XOR_verbosity = ("verbosity0", "verbosity1", "verbosity2")
    verbosity0 = traits.Bool(
        argstr="'-v0'", xor=_XOR_verbosity, desc="no messages will be reported"
    )
    verbosity1 = traits.Bool(
        argstr="'-v1'",
        xor=_XOR_verbosity,
        desc="messages will be reported but not the iteration-wise detailed messages",
    )
    verbosity2 = traits.Bool(
        argstr="'v2'",
        xor=_XOR_verbosity,
        desc="all the messages, including per-iteration, will be displayed",
    )
    shortMessages = traits.Bool(
        argstr="'-gui'", desc="Short messages instead of detailed messages"
    )
    displayModuleName = traits.Bool(
        argstr="'-m'", desc="Module name will be displayed in the messages"
    )
    displayTimestamps = traits.Bool(
        argstr="'-t'", desc="Timestamps will be displayed in the messages"
    )
    skipVolumetricProcessing = traits.Bool(
        argstr="'-S'",
        desc="Only surface registration and labeling will be performed. Volumetric "
        "processing will be skipped.",
    )
    useMultiThreading = traits.Bool(
        argstr="'-P'",
        desc="If multiple CPUs are present on the system, the code will try to use "
        "multithreading to make the execution fast.",
    )
    useSingleThreading = traits.Bool(argstr="'-U'", desc="Use single threaded mode.")


class SVReg(CommandLine):
    """
    surface and volume registration (svreg)
    This program registers a subject's BrainSuite-processed volume and surfaces
    to an atlas, allowing for automatic labelling of volume and surface ROIs.

    For more information, please see:
    http://brainsuite.org/processing/svreg/usage/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> svreg = brainsuite.SVReg()
    >>> svreg.inputs.subjectFilePrefix = 'home/user/btestsubject/testsubject'
    >>> svreg.inputs.refineOutputs = True
    >>> svreg.inputs.skipToVolumeReg = False
    >>> svreg.inputs. keepIntermediates = True
    >>> svreg.inputs.verbosity2 = True
    >>> svreg.inputs.displayTimestamps = True
    >>> svreg.inputs.useSingleThreading = True
    >>> results = svreg.run() #doctest: +SKIP


    """

    input_spec = SVRegInputSpec
    _cmd = "svreg.sh"

    def _format_arg(self, name, spec, value):
        if (
            name == "subjectFilePrefix"
            or name == "atlasFilePrefix"
            or name == "curveMatchingInstructions"
        ):
            return spec.argstr % os.path.expanduser(value)
        if name == "dataSinkDelay":
            return spec.argstr % ""
        return super(SVReg, self)._format_arg(name, spec, value)


class BDPInputSpec(CommandLineInputSpec):
    bfcFile = File(
        argstr="%s",
        mandatory=True,
        position=0,
        xor=["noStructuralRegistration"],
        desc="Specify absolute path to file produced by bfc. By default, bfc produces the file in "
        "the format: prefix.bfc.nii.gz",
    )
    noStructuralRegistration = traits.Bool(
        argstr="--no-structural-registration",
        mandatory=True,
        position=0,
        xor=["bfcFile"],
        desc="Allows BDP to work without any structural input. This can useful when "
        "one is only interested in diffusion modelling part of BDP. With this "
        "flag only fieldmap-based distortion correction is supported. "
        "outPrefix can be used to specify fileprefix of the output "
        "filenames. Change dwiMask to define region of interest "
        "for diffusion modelling.",
    )
    inputDiffusionData = File(
        argstr="--nii %s",
        mandatory=True,
        position=-2,
        desc="Specifies the absolute path and filename of the input diffusion data in 4D NIfTI-1 "
        "format. The flag must be followed by the filename. Only NIfTI-1 files "
        "with extension .nii or .nii.gz are supported. Furthermore, either  "
        "bMatrixFile, or a combination of both bValueFile and diffusionGradientFile "
        "must be used to provide the necessary b-matrices/b-values and gradient vectors. ",
    )
    bMatrixFile = File(
        argstr="--bmat %s",
        mandatory=True,
        xor=["BVecBValPair"],
        position=-1,
        desc="Specifies the absolute path and filename of the file containing b-matrices for "
        "diffusion-weighted scans. The flag must be followed by the filename. "
        "This file must be a plain text file containing 3x3 matrices for each "
        "diffusion encoding direction. It should contain zero matrices "
        'corresponding to b=0 images. This file usually has ".bmat" as its '
        "extension, and can be used to provide BDP with the more-accurate "
        "b-matrices as saved by some proprietary scanners. The b-matrices "
        "specified by the file must be in the voxel coordinates of the input "
        "diffusion weighted image (NIfTI file). In case b-matrices are not known/calculated, "
        "bvec and .bval files can be used instead (see diffusionGradientFile and bValueFile). ",
    )
    BVecBValPair = traits.List(
        traits.Str,
        minlen=2,
        maxlen=2,
        mandatory=True,
        position=-1,
        xor=["bMatrixFile"],
        argstr="--bvec %s --bval %s",
        desc="Must input a list containing first the BVector file, then the BValue file (both must be absolute paths)\n"
        "Example: bdp.inputs.BVecBValPair = ['/directory/subdir/prefix.dwi.bvec', '/directory/subdir/prefix.dwi.bval'] "
        "The first item in the list specifies the filename of the file containing b-values for the "
        "diffusion scan. The b-value file must be a plain-text file and usually has an "
        "extension of .bval\n"
        "The second item in the list specifies the filename of the file containing the diffusion gradient "
        "directions (specified in the voxel coordinates of the input "
        "diffusion-weighted image)The b-vectors file must be a plain text file and "
        "usually has an extension of .bvec ",
    )
    dataSinkDelay = traits.List(
        traits.Str,
        argstr="%s",
        desc="For use in parallel processing workflows including Brainsuite Cortical "
        "Surface Extraction sequence. Connect datasink out_file to dataSinkDelay "
        "to delay execution of BDP until dataSink has finished sinking outputs. "
        "In particular, BDP may be run after BFC has finished. For more information "
        "see http://brainsuite.org/processing/diffusion/pipeline/",
    )
    phaseEncodingDirection = traits.Enum(
        "x",
        "x-",
        "y",
        "y-",
        "z",
        "z-",
        argstr="--dir=%s",
        desc="Specifies the phase-encoding direction of the EPI (diffusion) images. "
        "It is same as the dominant direction of distortion in the images. This "
        "information is used to constrain the distortion correction along the "
        "specified direction. Directions are represented by any one of x, x-, y, "
        'y-, z or z-. "x" direction increases towards the right side of the '
        'subject, while "x-" increases towards the left side of the subject. '
        'Similarly, "y" and "y-" are along the anterior-posterior direction of '
        'the subject, and "z" & "z-" are along the inferior-superior direction. '
        'When this flag is not used, BDP uses "y" as the default phase-encoding '
        "direction. ",
    )
    echoSpacing = traits.Float(
        argstr="--echo-spacing=%f",
        desc="Sets the echo spacing to t seconds, which is used for fieldmap-based "
        "distortion correction. This flag is required when using fieldmapCorrection",
    )
    bValRatioThreshold = traits.Float(
        argstr="--bval-ratio-threshold %f",
        desc="Sets a threshold which is used to determine b=0 images. When there are "
        "no diffusion weighted image with b-value of zero, then BDP tries to use "
        "diffusion weighted images with a low b-value in place of b=0 image. The "
        "diffusion images with minimum b-value is used as b=0 image only if the "
        "ratio of the maximum and minimum b-value is more than the specified "
        "threshold. A lower value of threshold will allow diffusion images with "
        "higher b-value to be used as b=0 image. The default value of this "
        "threshold is set to 45, if this trait is not set. ",
    )
    estimateTensors = traits.Bool(
        argstr="--tensors",
        desc="Estimates diffusion tensors using a weighted log-linear estimation and "
        "saves derived diffusion tensor parameters (FA, MD, axial, radial, L2, "
        "L3). This is the default behavior if no diffusion modeling flags are "
        "specified. The estimated diffusion tensors can be visualized by loading "
        "the saved ``*.eig.nii.gz`` file in BrainSuite. BDP reports diffusivity (MD, "
        "axial, radial, L2 and L3) in a unit which is reciprocal inverse of the "
        "unit of input b-value. ",
    )
    estimateODF_FRACT = traits.Bool(
        argstr="--FRACT",
        desc="Estimates ODFs using the Funk-Radon and Cosine Transformation (FRACT). "
        'The outputs are saved in a separate directory with name "FRACT" and the '
        'ODFs can be visualized by loading the saved ".odf" file in BrainSuite. ',
    )
    estimateODF_FRT = traits.Bool(
        argstr="--FRT",
        desc="Estimates ODFs using Funk-Radon Transformation (FRT). The coefficient "
        'maps for ODFs are saved in a separate directory with name "FRT" and the '
        'ODFs can be visualized by loading the saved ".odf" file in BrainSuite. '
        "The derived generalized-FA (GFA) maps are also saved in the output "
        "directory. ",
    )
    estimateODF_3DShore = traits.Float(
        argstr="--3dshore --diffusion_time_ms %f",
        desc="Estimates ODFs using 3Dshore. Pass in diffusion time, in ms",
    )
    odfLambta = traits.Bool(
        argstr="--odf-lambda <L>",
        desc="Sets the regularization parameter, lambda, of the Laplace-Beltrami "
        "operator while estimating ODFs. The default value is set to 0.006 . This "
        "can be used to set the appropriate regularization for the input "
        "diffusion data. ",
    )
    t1Mask = File(
        argstr="--t1-mask %s",
        desc="Specifies the filename of the brain-mask file for input T1-weighted "
        "image. This mask can be same as the brain mask generated during "
        "BrainSuite extraction sequence. For best results, the mask should not "
        "include any extra-meningial tissues from T1-weighted image. The mask "
        "must be in the same coordinates as input T1-weighted image (i.e. should "
        "overlay correctly with input <fileprefix>.bfc.nii.gz file in "
        "BrainSuite). This mask is used for co-registration and defining brain "
        "boundary for statistics computation. The mask can be generated and/or "
        "edited in BrainSuite. In case outputDiffusionCoordinates is also "
        "used, this mask is first transformed to diffusion coordinate and the "
        "transformed mask is used for defining brain boundary in diffusion "
        "coordinates. When t1Mask is not set, BDP will try to use "
        "fileprefix>.mask.nii.gz as brain-mask. If <fileprefix>.mask.nii.gz is "
        "not found, then BDP will use the input <fileprefix>.bfc.nii.gz itself as "
        "mask (i.e. all non-zero voxels in <fileprefix>.bfc.nii.gz is assumed to "
        "constitute brain mask).  ",
    )
    dwiMask = File(
        argstr="--dwi-mask %s",
        desc="Specifies the filename of the brain-mask file for diffusion data. This "
        "mask is used only for co-registration purposes and can affect overall "
        "quality of co-registration (see t1Mask for definition of brain mask "
        "for statistics computation). The mask must be a 3D volume and should be "
        "in the same coordinates as input Diffusion file/data (i.e. should "
        "overlay correctly with input diffusion data in BrainSuite). For best "
        "results, the mask should include only brain voxels (CSF voxels around "
        "brain is also acceptable). When this flag is not used, BDP will generate "
        "a pseudo mask using first b=0 image volume and would save it as "
        "fileprefix>.dwi.RSA.mask.nii.gz. In case co-registration is not "
        "accurate with automatically generated pseudo mask, BDP should be re-run "
        "with a refined diffusion mask. The mask can be generated and/or edited "
        "in BrainSuite. ",
    )
    rigidRegMeasure = traits.Enum(
        "MI",
        "INVERSION",
        "BDP",
        argstr="--rigid-reg-measure %s",
        desc="Defines the similarity measure to be used for rigid registration. "
        'Possible measures are "MI", "INVERSION" and "BDP". MI measure uses '
        "normalized mutual information based cost function. INVERSION measure "
        "uses simpler cost function based on sum of squared difference by "
        "exploiting the approximate inverse-contrast relationship in T1- and "
        "T2-weighted images. BDP measure combines MI and INVERSION. It starts "
        "with INVERSION measure and refines the result with MI measure. BDP is "
        "the default measure when this trait is not set.  ",
    )
    dcorrRegMeasure = traits.Enum(
        "MI",
        "INVERSION-EPI",
        "INVERSION-T1",
        "INVERSION-BOTH",
        "BDP",
        argstr="--dcorr-reg-method %s",
        desc="Defines the method for registration-based distortion correction. "
        'Possible methods are "MI", "INVERSION-EPI", "INVERSION-T1", '
        'INVERSION-BOTH", and "BDP". MI method uses normalized mutual '
        "information based cost-function while estimating the distortion field. "
        "INVERSION-based method uses simpler cost function based on sum of "
        "squared difference by exploiting the known approximate contrast "
        "relationship in T1- and T2-weighted images. T2-weighted EPI is inverted "
        "when INVERSION-EPI is used; T1-image is inverted when INVERSION-T1 is "
        "used; and both are inverted when INVERSION-BOTH is used. BDP method add "
        "the MI-based refinement after the correction using INVERSION-BOTH "
        "method. BDP is the default method when this trait is not set.  ",
    )
    dcorrWeight = traits.Float(
        argstr="--dcorr-regularization-wt %f",
        desc="Sets the (scalar) weighting parameter for regularization penalty in "
        "registration-based distortion correction. Set this trait to a single, non-negative "
        "number which specifies the weight. A large regularization weight encourages "
        "smoother distortion field at the cost of low measure of image similarity "
        "after distortion correction. On the other hand, a smaller regularization "
        "weight can result into higher measure of image similarity but with "
        "unrealistic and unsmooth distortion field. A weight of 0.5 would reduce "
        "the penalty to half of the default regularization penalty (By default, this weight "
        "is set to 1.0). Similarly, a weight of 2.0 "
        "would increase the penalty to twice of the default penalty.  ",
    )
    skipDistortionCorr = traits.Bool(
        argstr="--no-distortion-correction",
        desc="Skips distortion correction completely and performs only a rigid "
        "registration of diffusion and T1-weighted image. This can be useful when "
        "the input diffusion images do not have any distortion or they have been "
        "corrected for distortion. ",
    )
    skipNonuniformityCorr = traits.Bool(
        argstr="--no-nonuniformity-correction",
        desc="Skips intensity non-uniformity correction in b=0 image for "
        "registration-based distortion correction. The intensity non-uniformity "
        "correction does not affect any diffusion modeling. ",
    )
    skipIntensityCorr = traits.Bool(
        argstr="--no-intensity-correction",
        xor=["fieldmapCorrectionMethod"],
        desc="Disables intensity correction when performing distortion correction. "
        "Intensity correction can change the noise distribution in the corrected "
        "image, but it does not affect estimated diffusion parameters like FA, "
        "etc. ",
    )
    fieldmapCorrection = File(
        argstr="--fieldmap-correction %s",
        requires=["echoSpacing"],
        desc="Use an acquired fieldmap for distortion correction. The fieldmap must "
        "have units of radians/second. Specify the filename of the fieldmap file. "
        "The field of view (FOV) of the fieldmap scan must cover the FOV of the diffusion "
        "scan. BDP will try to check the overlap of the FOV of the two scans and "
        'will issue a warning/error if the diffusion scan"s FOV is not fully '
        'covered by the fieldmap"s FOV. BDP uses all of the information saved in '
        "the NIfTI header to compute the FOV. If you get this error and think "
        "that it is incorrect, then it can be suppressed using the flag "
        "ignore-fieldmap-FOV. Neither the image matrix size nor the imaging "
        "grid resolution of the fieldmap needs to be the same as that of the "
        "diffusion scan, but the fieldmap must be pre-registred to the diffusion "
        "scan. BDP does NOT align the fieldmap to the diffusion scan, nor does it "
        "check the alignment of the fieldmap and diffusion scans. Only NIfTI "
        "files with extension of .nii or .nii.gz are supported. Fieldmap-based "
        "distortion correction also requires the echoSpacing. Also "
        "fieldmapCorrectionMethod allows you to define method for "
        "distortion correction. least squares is the default method. ",
    )
    fieldmapCorrectionMethod = traits.Enum(
        "pixelshift",
        "leastsq",
        xor=["skipIntensityCorr"],
        argstr="--fieldmap-correction-method %s",
        desc="Defines the distortion correction method while using fieldmap. "
        'Possible methods are "pixelshift" and "leastsq". leastsq is the default '
        "method when this flag is not used. Pixel-shift (pixelshift) method uses "
        "image interpolation to un-distort the distorted diffusion images. Least "
        "squares (leastsq) method uses a physical model of distortion which is "
        "more accurate (and more computationally expensive) than pixel-shift "
        "method.",
    )
    ignoreFieldmapFOV = traits.Bool(
        argstr="--ignore-fieldmap-fov",
        desc="Supresses the error generated by an insufficient field of view of the "
        "input fieldmap and continues with the processing. It is useful only when "
        "used with fieldmap-based distortion correction. See "
        "fieldmap-correction for a detailed explanation. ",
    )
    fieldmapSmooth = traits.Float(
        argstr="--fieldmap-smooth3=%f",
        desc="Applies 3D Gaussian smoothing with a standard deviation of S "
        "millimeters (mm) to the input fieldmap before applying distortion "
        "correction. This trait is only useful with "
        "fieldmapCorrection. Skip this trait for no smoothing. ",
    )
    transformDiffusionVolume = File(
        argstr="--transform-diffusion-volume %s",
        desc="This flag allows to define custom volumes in diffusion coordinate "
        "which would be transformed into T1 coordinate in a rigid fashion. The "
        "flag must be followed by the name of either a NIfTI file or of a folder "
        "that contains one or more NIfTI files. All of the files must be in "
        "diffusion coordinate, i.e. the files should overlay correctly with the "
        "diffusion scan in BrainSuite. Only NIfTI files with an extension of .nii "
        "or .nii.gz are supported. The transformed files are written to the "
        'output directory with suffix ".T1_coord" in the filename and will not be '
        "corrected for distortion, if any. The trait transformInterpolation can "
        "be used to define the type of interpolation that would be used (default "
        "is set to linear). If you are attempting to transform a label file or "
        'mask file, use "nearest" interpolation method with transformInterpolation. '
        "See also transformT1Volume and transformInterpolation",
    )
    transformT1Volume = File(
        argstr="--transform-t1-volume %s",
        desc="Same as transformDiffusionVolume except that files specified must "
        "be in T1 coordinate, i.e. the files should overlay correctly with the "
        "input <fileprefix>.bfc.nii.gz files in BrainSuite. BDP transforms these "
        "data/images from T1 coordinate to diffusion coordinate. The transformed "
        'files are written to the output directory with suffix ".D_coord" in the '
        "filename. See also transformDiffusionVolume  and transformInterpolation. ",
    )
    transformInterpolation = traits.Enum(
        "linear",
        "nearest",
        "cubic",
        "spline",
        argstr="--transform-interpolation %s",
        desc="Defines the type of interpolation method which would be used while "
        "transforming volumes defined by transformT1Volume and "
        'transformDiffusionVolume. Possible methods are "linear", "nearest", '
        '"cubic" and "spline". By default, "linear" interpolation is used. ',
    )
    transformT1Surface = File(
        argstr="--transform-t1-surface %s",
        desc="Similar to transformT1Volume, except that this flag allows "
        "transforming surfaces (instead of volumes) in T1 coordinate into "
        "diffusion coordinate in a rigid fashion. The flag must be followed by "
        "the name of either a .dfs file or of a folder that contains one or more "
        "dfs files. All of the files must be in T1 coordinate, i.e. the files "
        "should overlay correctly with the T1-weighted scan in BrainSuite. The "
        "transformed files are written to the output directory with suffix "
        'D_coord" in the filename. ',
    )
    transformDiffusionSurface = File(
        argstr="--transform-diffusion-surface %s",
        desc="Same as transformT1Volume, except that the .dfs files specified "
        "must be in diffusion coordinate, i.e. the surface files should overlay "
        "correctly with the diffusion scan in BrainSuite. The transformed files "
        'are written to the output directory with suffix ".T1_coord" in the '
        "filename. See also transformT1Volume. ",
    )
    transformDataOnly = traits.Bool(
        argstr="--transform-data-only",
        desc="Skip all of the processing (co-registration, distortion correction and "
        "tensor/ODF estimation) and directly start transformation of defined "
        "custom volumes, mask and labels (using transformT1Volume, "
        "transformDiffusionVolume, transformT1Surface, "
        "transformDiffusionSurface, customDiffusionLabel, "
        "customT1Label). This flag is useful when BDP was previously run on a "
        "subject (or <fileprefix>) and some more data (volumes, mask or labels) "
        "need to be transformed across the T1-diffusion coordinate spaces. This "
        "assumes that all the necessary files were generated earlier and all of "
        "the other flags MUST be used in the same way as they were in the initial "
        "BDP run that processed the data. ",
    )
    generateStats = traits.Bool(
        argstr="--generate-stats",
        desc="Generate ROI-wise statistics of estimated diffusion tensor parameters. "
        "Units of the reported statistics are same as that of the estimated "
        "tensor parameters (see estimateTensors). Mean, variance, and voxel counts of "
        "white matter(WM), grey matter(GM), and both WM and GM combined are "
        "written for each estimated parameter in a separate comma-seperated value "
        "csv) file. BDP uses the ROI labels generated by Surface-Volume "
        "Registration (SVReg) in the BrainSuite extraction sequence. "
        "Specifically, it looks for labels saved in either "
        "fileprefix>.svreg.corr.label.nii.gz or <fileprefix>.svreg.label.nii.gz. "
        "In case both files are present, only the first file is used. Also see "
        "customDiffusionLabel and customT1Label for specifying your own "
        "ROIs. It is also possible to forgo computing the SVReg ROI-wise "
        "statistics and only compute stats with custom labels if SVReg label is "
        "missing. BDP also transfers (and saves) the label/mask files to "
        "appropriate coordinates before computing statistics. Also see "
        "outputDiffusionCoordinates for outputs in diffusion coordinate and "
        "forcePartialROIStats for an important note about field of view of "
        "diffusion and T1-weighted scans. ",
    )
    onlyStats = traits.Bool(
        argstr="--generate-only-stats",
        desc="Skip all of the processing (co-registration, distortion correction and "
        "tensor/ODF estimation) and directly start computation of statistics. "
        "This flag is useful when BDP was previously run on a subject (or "
        "fileprefix>) and statistics need to be (re-)computed later. This "
        "assumes that all the necessary files were generated earlier. All of the "
        "other flags MUST be used in the same way as they were in the initial BDP "
        "run that processed the data. ",
    )
    forcePartialROIStats = traits.Bool(
        argstr="--force-partial-roi-stats",
        desc="The field of view (FOV) of the diffusion and T1-weighted scans may "
        "differ significantly in some situations. This may result in partial "
        "acquisitions of some ROIs in the diffusion scan. By default, BDP does "
        "not compute statistics for partially acquired ROIs and shows warnings. "
        "This flag forces computation of statistics for all ROIs, including those "
        "which are partially acquired. When this flag is used, number of missing "
        "voxels are also reported for each ROI in statistics files. Number of "
        "missing voxels are reported in the same coordinate system as the "
        "statistics file.  ",
    )
    customDiffusionLabel = File(
        argstr="--custom-diffusion-label %s",
        desc="BDP supports custom ROIs in addition to those generated by BrainSuite "
        "SVReg) for ROI-wise statistics calculation. The flag must be followed "
        "by the name of either a file (custom ROI file) or of a folder that "
        "contains one or more ROI files. All of the files must be in diffusion "
        "coordinate, i.e. the label files should overlay correctly with the "
        "diffusion scan in BrainSuite. These input label files are also "
        "transferred (and saved) to T1 coordinate for statistics in T1 "
        "coordinate. BDP uses nearest-neighborhood interpolation for this "
        "transformation. Only NIfTI files, with an extension of .nii or .nii.gz "
        "are supported. In order to avoid confusion with other ROI IDs in the "
        "statistic files, a 5-digit ROI ID is generated for each custom label "
        "found and the mapping of ID to label file is saved in the file "
        "fileprefix>.BDP_ROI_MAP.xml. Custom label files can also be generated "
        "by using the label painter tool in BrainSuite. See also "
        "customLabelXML",
    )
    customT1Label = File(
        argstr="--custom-t1-label %s",
        desc="Same as customDiffusionLabelexcept that the label files specified "
        "must be in T1 coordinate, i.e. the label files should overlay correctly "
        "with the T1-weighted scan in BrainSuite. If the trait "
        "outputDiffusionCoordinates is also used then these input label files "
        "are also transferred (and saved) to diffusion coordinate for statistics "
        "in diffusion coordinate. BDP uses nearest-neighborhood interpolation for "
        "this transformation. See also customLabelXML. ",
    )
    customLabelXML = File(
        argstr="--custom-label-xml %s",
        desc="BrainSuite saves a descriptions of the SVReg labels (ROI name, ID, "
        "color, and description) in an .xml file "
        'brainsuite_labeldescription.xml). BDP uses the ROI ID"s from this xml '
        "file to report statistics. This flag allows for the use of a custom "
        "label description xml file. The flag must be followed by an xml "
        "filename. This can be useful when you want to limit the ROIs for which "
        "you compute statistics. You can also use custom xml files to name your "
        'own ROIs (assign ID"s) for custom labels. BrainSuite can save a label '
        "description in .xml format after using the label painter tool to create "
        'a ROI label. The xml file MUST be in the same format as BrainSuite"s '
        "label description file (see brainsuite_labeldescription.xml for an "
        "example). When this flag is used, NO 5-digit ROI ID is generated for "
        "custom label files and NO Statistics will be calculated for ROIs not "
        "identified in the custom xml file. See also customDiffusionLabel and "
        "customT1Label.",
    )
    outputSubdir = traits.Str(
        argstr="--output-subdir %s",
        desc="By default, BDP writes out all the output (and intermediate) files in "
        "the same directory (or folder) as the BFC file. This flag allows to "
        "specify a sub-directory name in which output (and intermediate) files "
        "would be written. BDP will create the sub-directory in the same "
        "directory as BFC file. <directory_name> should be the name of the "
        "sub-directory without any path. This can be useful to organize all "
        "outputs generated by BDP in a separate sub-directory. ",
    )
    outputDiffusionCoordinates = traits.Bool(
        argstr="--output-diffusion-coordinate",
        desc="Enables estimation of diffusion tensors and/or ODFs (and statistics if "
        "applicable) in the native diffusion coordinate in addition to the "
        "default T1-coordinate. All native diffusion coordinate files are saved "
        'in a seperate folder named "diffusion_coord_outputs". In case statistics '
        "computation is required, it will also transform/save all label/mask "
        "files required to diffusion coordinate (see generateStats for "
        "details). ",
    )
    flagConfigFile = File(
        argstr="--flag-conf-file %s",
        desc="Uses the defined file to specify BDP flags which can be useful for "
        "batch processing. A flag configuration file is a plain text file which "
        'can contain any number of BDP"s optional flags (and their parameters) '
        "separated by whitespace. Everything coming after # until end-of-line is "
        "treated as comment and is ignored. If a flag is defined in configuration "
        "file and is also specified in the command used to run BDP, then the "
        "later get preference and overrides the definition in configuration "
        "file. ",
    )
    outPrefix = traits.Str(
        argstr="--output-fileprefix %s",
        desc="Specifies output fileprefix when noStructuralRegistration is "
        "used. The fileprefix can not start with a dash (-) and should be a "
        "simple string reflecting the absolute path to desired location, along with outPrefix. When this flag is "
        "not specified (and noStructuralRegistration is used) then the output "
        "files have same file-base as the input diffusion file. This trait is "
        "ignored when noStructuralRegistration is not used. ",
    )
    threads = traits.Int(
        argstr="--threads=%d",
        desc="Sets the number of parallel process threads which can be used for "
        "computations to N, where N must be an integer. Default value of N is "
        " ",
    )
    lowMemory = traits.Bool(
        argstr="--low-memory",
        desc="Activates low-memory mode. This will run the registration-based "
        "distortion correction at a lower resolution, which could result in a "
        "less-accurate correction. This should only be used when no other "
        "alternative is available. ",
    )
    ignoreMemory = traits.Bool(
        argstr="--ignore-memory",
        desc="Deactivates the inbuilt memory checks and forces BDP to run "
        "registration-based distortion correction at its default resolution even "
        "on machines with a low amount of memory. This may result in an "
        "out-of-memory error when BDP cannot allocate sufficient memory. ",
    )


class BDP(CommandLine):
    """
    BrainSuite Diffusion Pipeline (BDP) enables fusion of diffusion and
    structural MRI information for advanced image and connectivity analysis.
    It provides various methods for distortion correction, co-registration,
    diffusion modeling (DTI and ODF) and basic ROI-wise statistic. BDP is a
    flexible and diverse tool which supports wide variety of diffusion
    datasets.
    For more information, please see:

    http://brainsuite.org/processing/diffusion/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> bdp = brainsuite.BDP()
    >>> bdp.inputs.bfcFile = '/directory/subdir/prefix.bfc.nii.gz'
    >>> bdp.inputs.inputDiffusionData = '/directory/subdir/prefix.dwi.nii.gz'
    >>> bdp.inputs.BVecBValPair = ['/directory/subdir/prefix.dwi.bvec', '/directory/subdir/prefix.dwi.bval']
    >>> results = bdp.run() #doctest: +SKIP


    """

    input_spec = BDPInputSpec
    _cmd = "bdp.sh"

    def _format_arg(self, name, spec, value):
        if name == "BVecBValPair":
            return spec.argstr % (value[0], value[1])
        if name == "dataSinkDelay":
            return spec.argstr % ""
        return super(BDP, self)._format_arg(name, spec, value)


class ThicknessPVCInputSpec(CommandLineInputSpec):
    subjectFilePrefix = traits.Str(
        argstr="%s",
        mandatory=True,
        desc="Absolute path and filename prefix of the subject data",
    )


class ThicknessPVC(CommandLine):
    """
    ThicknessPVC computes cortical thickness using partial tissue fractions.
    This thickness measure is then transferred to the atlas surface to
    facilitate population studies. It also stores the computed thickness into
    separate hemisphere files and subject thickness mapped to the atlas
    hemisphere surfaces. ThicknessPVC is not run through the main SVReg
    sequence, and should be used after executing the BrainSuite and SVReg
    sequence.
    For more informaction, please see:

    http://brainsuite.org/processing/svreg/svreg_modules/

    Examples
    --------

    >>> from nipype.interfaces import brainsuite
    >>> thicknessPVC = brainsuite.ThicknessPVC()
    >>> thicknessPVC.inputs.subjectFilePrefix = 'home/user/btestsubject/testsubject'
    >>> results = thicknessPVC.run() #doctest: +SKIP

    """

    input_spec = ThicknessPVCInputSpec
    _cmd = "thicknessPVC.sh"


# used to generate file names for outputs
# removes pathway and extension of inputName, returns concatenation of:
# inputName and suffix
def getFileName(inputName, suffix):
    fullInput = os.path.basename(inputName)
    dotRegex = regex.compile("[^.]+")
    # extract between last slash and first period
    inputNoExtension = dotRegex.findall(fullInput)[0]
    return os.path.abspath("".join((inputNoExtension, suffix)))


def l_outputs(self):
    outputs = self.output_spec().get()
    for key in outputs:
        name = self._gen_filename(key)
        if name is not None:
            outputs[key] = name

    return outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import TProject


def test_TProject_inputs():
    input_map = dict(
        TR=dict(
            argstr="-TR %g",
        ),
        args=dict(
            argstr="%s",
        ),
        automask=dict(
            argstr="-automask",
            xor=["mask"],
        ),
        bandpass=dict(
            argstr="-bandpass %g %g",
        ),
        blur=dict(
            argstr="-blur %g",
        ),
        cenmode=dict(
            argstr="-cenmode %s",
        ),
        censor=dict(
            argstr="-censor %s",
            extensions=None,
        ),
        censortr=dict(
            argstr="-CENSORTR %s",
        ),
        concat=dict(
            argstr="-concat %s",
            extensions=None,
        ),
        dsort=dict(
            argstr="-dsort %s...",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-input %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=1,
        ),
        mask=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        noblock=dict(
            argstr="-noblock",
        ),
        norm=dict(
            argstr="-norm",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        ort=dict(
            argstr="-ort %s",
            extensions=None,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            name_source="in_file",
            name_template="%s_tproject",
            position=-1,
        ),
        outputtype=dict(),
        polort=dict(
            argstr="-polort %d",
        ),
        stopband=dict(
            argstr="-stopband %g %g",
        ),
    )
    inputs = TProject.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TProject_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = TProject.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Camino top level namespace
"""

from .connectivity import Conmat
from .convert import (
    Image2Voxel,
    FSL2Scheme,
    VtkStreamlines,
    ProcStreamlines,
    TractShredder,
    DT2NIfTI,
    NIfTIDT2Camino,
    AnalyzeHeader,
    Shredder,
)
from .dti import (
    DTIFit,
    ModelFit,
    DTLUTGen,
    PicoPDFs,
    Track,
    TrackPICo,
    TrackBayesDirac,
    TrackDT,
    TrackBallStick,
    TrackBootstrap,
    TrackBedpostxDeter,
    TrackBedpostxProba,
    ComputeFractionalAnisotropy,
    ComputeMeanDiffusivity,
    ComputeTensorTrace,
    ComputeEigensystem,
    DTMetric,
)
from .calib import SFPICOCalibData, SFLUTGen
from .odf import QBallMX, LinRecon, SFPeaks, MESD
from .utils import ImageStats



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import SurfaceSmooth


def test_SurfaceSmooth_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        cortex=dict(
            argstr="--cortex",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fwhm=dict(
            argstr="--fwhm %.4f",
            xor=["smooth_iters"],
        ),
        hemi=dict(
            argstr="--hemi %s",
            mandatory=True,
        ),
        in_file=dict(
            argstr="--sval %s",
            extensions=None,
            mandatory=True,
        ),
        out_file=dict(
            argstr="--tval %s",
            extensions=None,
            genfile=True,
        ),
        reshape=dict(
            argstr="--reshape",
        ),
        smooth_iters=dict(
            argstr="--smooth %d",
            xor=["fwhm"],
        ),
        subject_id=dict(
            argstr="--s %s",
            mandatory=True,
        ),
        subjects_dir=dict(),
    )
    inputs = SurfaceSmooth.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SurfaceSmooth_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = SurfaceSmooth.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Camino-Trackvis allows interoperability between Camino and TrackVis."""
from .convert import Camino2Trackvis, Trackvis2Camino



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..em import EM


def test_EM_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bc_order_val=dict(
            argstr="-bc_order %s",
            usedefault=True,
        ),
        bc_thresh_val=dict(
            argstr="-bc_thresh %s",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-in %s",
            extensions=None,
            mandatory=True,
            position=4,
        ),
        mask_file=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        max_iter=dict(
            argstr="-max_iter %s",
            usedefault=True,
        ),
        min_iter=dict(
            argstr="-min_iter %s",
            usedefault=True,
        ),
        mrf_beta_val=dict(
            argstr="-mrf_beta %s",
        ),
        no_prior=dict(
            argstr="-nopriors %s",
            mandatory=True,
            xor=["prior_4D", "priors"],
        ),
        out_bc_file=dict(
            argstr="-bc_out %s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_bc_em.nii.gz",
        ),
        out_file=dict(
            argstr="-out %s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_em.nii.gz",
        ),
        out_outlier_file=dict(
            argstr="-out_outlier %s",
            extensions=None,
            name_source=["in_file"],
            name_template="%s_outlier_em.nii.gz",
        ),
        outlier_val=dict(
            argstr="-outlier %s %s",
        ),
        prior_4D=dict(
            argstr="-prior4D %s",
            extensions=None,
            mandatory=True,
            xor=["no_prior", "priors"],
        ),
        priors=dict(
            argstr="%s",
            mandatory=True,
            xor=["no_prior", "prior_4D"],
        ),
        reg_val=dict(
            argstr="-reg %s",
        ),
        relax_priors=dict(
            argstr="-rf %s %s",
        ),
    )
    inputs = EM.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_EM_outputs():
    output_map = dict(
        out_bc_file=dict(
            extensions=None,
        ),
        out_file=dict(
            extensions=None,
        ),
        out_outlier_file=dict(
            extensions=None,
        ),
    )
    outputs = EM.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import pytest
from ..base import EngineBase
from ....interfaces import base as nib
from ....interfaces import utility as niu
from ... import engine as pe


class InputSpec(nib.TraitedSpec):
    input1 = nib.traits.Int(desc="a random int")
    input2 = nib.traits.Int(desc="a random int")
    input_file = nib.File(desc="Random File")


class OutputSpec(nib.TraitedSpec):
    output1 = nib.traits.List(nib.traits.Int, desc="outputs")


class EngineTestInterface(nib.SimpleInterface):
    input_spec = InputSpec
    output_spec = OutputSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        self._results["output1"] = [1, self.inputs.input1]
        return runtime


@pytest.mark.parametrize("name", ["valid1", "valid_node", "valid-node", "ValidNode0"])
def test_create(name):
    base = EngineBase(name=name)
    assert base.name == name


@pytest.mark.parametrize(
    "name", ["invalid*1", "invalid.1", "invalid@", "in/valid", None]
)
def test_create_invalid(name):
    with pytest.raises(ValueError):
        EngineBase(name=name)


def test_hierarchy():
    base = EngineBase(name="nodename")
    base._hierarchy = "some.history.behind"

    assert base.name == "nodename"
    assert base.fullname == "some.history.behind.nodename"


def test_clone():
    base = EngineBase(name="nodename")
    base2 = base.clone("newnodename")

    assert (
        base.base_dir == base2.base_dir
        and base.config == base2.config
        and base2.name == "newnodename"
    )

    with pytest.raises(ValueError):
        base.clone("nodename")


def test_clone_node_iterables(tmpdir):
    tmpdir.chdir()

    def addstr(string):
        return "%s + 2" % string

    subject_list = ["sub-001", "sub-002"]
    inputnode = pe.Node(niu.IdentityInterface(fields=["subject"]), name="inputnode")
    inputnode.iterables = [("subject", subject_list)]

    node_1 = pe.Node(
        niu.Function(input_names="string", output_names="string", function=addstr),
        name="node_1",
    )
    node_2 = node_1.clone("node_2")

    workflow = pe.Workflow(name="iter_clone_wf")
    workflow.connect(
        [
            (inputnode, node_1, [("subject", "string")]),
            (node_1, node_2, [("string", "string")]),
        ]
    )
    workflow.run()



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import Zeropad


def test_Zeropad_inputs():
    input_map = dict(
        A=dict(
            argstr="-A %i",
            xor=["master"],
        ),
        AP=dict(
            argstr="-AP %i",
            xor=["master"],
        ),
        I=dict(
            argstr="-I %i",
            xor=["master"],
        ),
        IS=dict(
            argstr="-IS %i",
            xor=["master"],
        ),
        L=dict(
            argstr="-L %i",
            xor=["master"],
        ),
        P=dict(
            argstr="-P %i",
            xor=["master"],
        ),
        R=dict(
            argstr="-R %i",
            xor=["master"],
        ),
        RL=dict(
            argstr="-RL %i",
            xor=["master"],
        ),
        S=dict(
            argstr="-S %i",
            xor=["master"],
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_files=dict(
            argstr="%s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=-1,
        ),
        master=dict(
            argstr="-master %s",
            extensions=None,
            xor=["I", "S", "A", "P", "L", "R", "z", "RL", "AP", "IS", "mm"],
        ),
        mm=dict(
            argstr="-mm",
            xor=["master"],
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            name_template="zeropad",
        ),
        outputtype=dict(),
        z=dict(
            argstr="-z %i",
            xor=["master"],
        ),
    )
    inputs = Zeropad.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Zeropad_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = Zeropad.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value
